from decimal import Decimal
from itertools import combinations
from bloqade.atom_arrangement import Chain
from bloqade.codegen.emulator_ir import EmulatorProgramCodeGen
from bloqade.emulate.codegen.hamiltonian import CompileCache, RydbergHamiltonianCodeGen
from functools import reduce
import numpy as np
import pytest

np.random.seed(2304023)
L_VALUES = [1, 2, 3, 4, 5]


def get_manybody_op(i, L, op):
    if op.ndim == 1:
        Ident = np.ones_like(op)
    else:
        Ident = np.eye(op.shape[0])

    if i < 0 or i >= L:
        raise ValueError("i must be in range [0, L-1]")

    if L <= 1:
        return op

    i = L - i - 1

    if i == 0:
        rhs = reduce(np.kron, [Ident] * (L - 1))
        return np.kron(op, rhs)
    elif i == L - 1:
        lhs = reduce(np.kron, [Ident] * i)
        return np.kron(lhs, op)
    else:
        lhs = reduce(np.kron, [Ident] * i)
        rhs = reduce(np.kron, [Ident] * (L - i - 1))
        return np.kron(np.kron(lhs, op), rhs)


def project_to_subspace(operator, configurations):
    from scipy.sparse import csr_matrix

    proj_shape = (configurations.size, operator.shape[0])
    data = np.ones_like(configurations)
    rows = np.arange(configurations.size)
    cols = configurations
    proj = csr_matrix((data, (rows, cols)), shape=proj_shape)

    if operator.ndim == 1:
        return proj @ operator
    else:
        return proj @ operator @ proj.T


####################
# Test uniform ops #
####################


@pytest.mark.parametrize("L", L_VALUES)
def test_2_level_uniform_detuningl(L):
    circuit = (
        Chain(L, lattice_spacing=6.1).rydberg.detuning.uniform.constant(1.0, 1.0)
    ).parse_circuit()

    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)
    detuning_op = np.array([0, -1], dtype=int)

    detuning = sum([get_manybody_op(i, L, detuning_op) for i in range(L)])

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    detuning_op_proj = project_to_subspace(detuning, hamiltonian.space.configurations)

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning_op_proj)


@pytest.mark.parametrize("L", L_VALUES)
def test_2_level_uniform_rabi_real(L):
    circuit = (
        Chain(L, lattice_spacing=6.1).rydberg.rabi.amplitude.uniform.constant(1.0, 1.0)
    ).parse_circuit()

    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)
    rabi_op = np.array([[0, 1], [1, 0]], dtype=int)

    rabi = sum([get_manybody_op(i, L, rabi_op) for i in range(L)])

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


@pytest.mark.parametrize("L", L_VALUES)
def test_2_level_uniform_complex(L):
    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.uniform.constant(1.0, 1.0)
        .phase.uniform.constant(0.0, 1.0)
    ).parse_circuit()

    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)
    rabi_op = np.array([[0, 0], [1, 0]], dtype=int)

    rabi = sum([get_manybody_op(i, L, rabi_op) for i in range(L)])

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


@pytest.mark.parametrize("L", L_VALUES)
def test_3_level_uniform_detuning(L):
    circuit = (
        Chain(L, lattice_spacing=6.1).rydberg.detuning.uniform.constant(1.0, 1.0)
    ).parse_circuit()

    emu_prog = EmulatorProgramCodeGen(use_hyperfine=True).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)
    detuning_op = np.array([0, 0, -1], dtype=int)

    detuning = sum([get_manybody_op(i, L, detuning_op) for i in range(L)])

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2, use_hyperfine=True).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    detuning_op_proj = project_to_subspace(detuning, hamiltonian.space.configurations)

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning_op_proj)

    # check hyperfine detuning value

    circuit = (
        Chain(L, lattice_spacing=6.1).hyperfine.detuning.uniform.constant(1.0, 1.0)
    ).parse_circuit()

    emu_prog = EmulatorProgramCodeGen(use_hyperfine=True).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)
    detuning_op = np.array([0, -1, 0], dtype=int)

    detuning = sum([get_manybody_op(i, L, detuning_op) for i in range(L)])

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    detuning_op_proj = project_to_subspace(detuning, hamiltonian.space.configurations)

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning_op_proj)


@pytest.mark.parametrize("L", L_VALUES)
def test_3_level_uniform_rabi_complex(L):
    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.uniform.constant(1.0, 1.0)
        .phase.uniform.constant(0.0, 1.0)
    ).parse_circuit()
    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(use_hyperfine=True).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0, 0], [0, 0, 0], [0, 1, 0]], dtype=int)

    rabi = sum([get_manybody_op(i, L, rabi_op) for i in range(L)])

    print(rabi)
    print(hamiltonian.rabi_ops[0].op.tocsr().toarray())

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2, use_hyperfine=True).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)

    # check hyperfine rabi value

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .hyperfine.rabi.amplitude.uniform.constant(1.0, 1.0)
        .phase.uniform.constant(0.0, 1.0)
    ).parse_circuit()
    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0, 0], [1, 0, 0], [0, 0, 0]], dtype=int)

    rabi = sum([get_manybody_op(i, L, rabi_op) for i in range(L)])

    print(rabi)
    print(hamiltonian.rabi_ops[0].op.tocsr().toarray())

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


####################
# Test single atom #
####################


@pytest.mark.parametrize(("i", "L"), [(i, L) for L in L_VALUES for i in range(L)])
def test_2_level_single_atom_detuning(i, L):
    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.detuning.location(i, 2.0)
        .constant(1.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    detuning_op = np.array([0, -1], dtype=int)

    detuning = 2.0 * get_manybody_op(i, L, detuning_op)

    assert np.all(hamiltonian.detuning_ops[0].get_diagonal(0.0) == detuning)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    detuning_op_proj = project_to_subspace(detuning, hamiltonian.space.configurations)

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning_op_proj)


@pytest.mark.parametrize(("i", "L"), [(i, L) for L in L_VALUES for i in range(L)])
def test_2_level_single_atom_rabi_real(i, L):
    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.location(i, 0.5)
        .constant(1.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 1], [1, 0]], dtype=int)

    rabi = get_manybody_op(i, L, rabi_op)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


@pytest.mark.parametrize(("i", "L"), [(i, L) for L in L_VALUES for i in range(L)])
def test_2_level_single_atom_complex(i, L):
    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.location(i, 0.5)
        .constant(1.0, 1.0)
        .phase.location(i)
        .constant(0.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0], [1, 0]], dtype=int)
    rabi = get_manybody_op(i, L, rabi_op)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


@pytest.mark.parametrize(("i", "L"), [(i, L) for L in L_VALUES for i in range(L)])
def test_3_level_single_atom_detuning(i, L):
    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.detuning.location(i, 2.0)
        .constant(1.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(use_hyperfine=True).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    detuning_op = np.array([0, 0, -1], dtype=int)

    detuning = 2.0 * get_manybody_op(i, L, detuning_op)

    assert np.all(hamiltonian.detuning_ops[0].get_diagonal(0.0) == detuning)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2, use_hyperfine=True).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    detuning_op_proj = project_to_subspace(detuning, hamiltonian.space.configurations)

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning_op_proj)

    # check hyperfine detuning value

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .hyperfine.detuning.location(i, 2.0)
        .constant(1.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    detuning_op = np.array([0, -1, 0], dtype=int)

    detuning = 2.0 * get_manybody_op(i, L, detuning_op)

    assert np.all(hamiltonian.detuning_ops[0].get_diagonal(0.0) == detuning)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    detuning_op_proj = project_to_subspace(detuning, hamiltonian.space.configurations)

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning_op_proj)


@pytest.mark.parametrize(("i", "L"), [(i, L) for L in L_VALUES for i in range(L)])
def test_3_level_single_atom_rabi_real(i, L):
    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.location(i, 0.5)
        .constant(1.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(use_hyperfine=True).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0]], dtype=int)

    rabi = get_manybody_op(i, L, rabi_op)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2, use_hyperfine=True).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)

    # check hyperfine rabi value

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .hyperfine.rabi.amplitude.location(i, 0.5)
        .constant(1.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]], dtype=int)

    rabi = get_manybody_op(i, L, rabi_op)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


@pytest.mark.parametrize(
    ("i", "L"), [(i, L) for L in [1, 2, 3, 4, 5, 6] for i in range(L)]
)
def test_3_level_single_atom_rabi_complex(i, L):
    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.location(i, 0.5)
        .constant(1.0, 1.0)
        .phase.location(i)
        .constant(0.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(use_hyperfine=True).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0, 0], [0, 0, 0], [0, 1, 0]], dtype=int)
    rabi = get_manybody_op(i, L, rabi_op)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2, use_hyperfine=True).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)

    # check hyperfine rabi value

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .hyperfine.rabi.amplitude.location(i, 0.5)
        .constant(1.0, 1.0)
        .phase.location(i)
        .constant(0.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0, 0], [1, 0, 0], [0, 0, 0]], dtype=int)
    rabi = get_manybody_op(i, L, rabi_op)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


####################
# Test full mask   #
####################


@pytest.mark.parametrize("L", L_VALUES)
def test_2_level_mask_detuning(L):
    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.detuning.scale("detuning_mask")
        .constant(1.0, 1.0)
    ).parse_circuit()

    detuning_mask = [Decimal(str(np.random.normal(0.0, 1.0))) for _ in range(L)]
    mask_assignments = dict(detuning_mask=detuning_mask)

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(mask_assignments).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    detuning_op = np.array([0, -1], dtype=int)

    detuning = sum(
        float(mask) * get_manybody_op(i, L, detuning_op)
        for i, mask in enumerate(detuning_mask)
    )

    assert np.all(hamiltonian.detuning_ops[0].get_diagonal(0.0) == detuning)

    emu_prog = EmulatorProgramCodeGen(mask_assignments, blockade_radius=6.2).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    detuning_op_proj = project_to_subspace(detuning, hamiltonian.space.configurations)

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning_op_proj)


@pytest.mark.parametrize("L", L_VALUES)
def test_2_level_mask_real(L):
    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.scale("rabi_mask")
        .constant(1.0, 1.0)
    ).parse_circuit()

    rabi_mask = [Decimal(str(np.random.normal(0.0, 1.0))) for _ in range(L)]
    mask_assignments = dict(rabi_mask=rabi_mask)

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(mask_assignments).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 1], [1, 0]], dtype=int)

    if L == 1:
        rabi = get_manybody_op(0, L, rabi_op)
    else:
        rabi = sum(
            float(mask) * get_manybody_op(i, L, rabi_op)
            for i, mask in enumerate(rabi_mask)
        )

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(mask_assignments, blockade_radius=6.2).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


@pytest.mark.parametrize("L", L_VALUES)
def test_2_level_mask_complex(L):
    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.scale("rabi_mask")
        .constant(1.0, 1.0)
        .phase.uniform.constant(0.0, 1.0)
    ).parse_circuit()

    rabi_mask = [Decimal(str(np.random.normal(0.0, 1.0))) for _ in range(L)]
    mask_assignments = dict(
        rabi_mask=rabi_mask,
    )

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(mask_assignments).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0], [1, 0]], dtype=int)

    if L == 1:
        rabi = get_manybody_op(0, L, rabi_op)
    else:
        rabi = sum(
            float(mask) * get_manybody_op(i, L, rabi_op)
            for i, mask in enumerate(rabi_mask)
        )

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(mask_assignments, blockade_radius=6.2).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


@pytest.mark.parametrize("L", L_VALUES)
def test_3_level_detuning(L):
    detuning_mask = [Decimal(str(np.random.normal(0.0, 1.0))) for _ in range(L)]
    mask_assignments = dict(detuning_mask=detuning_mask)

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.detuning.scale("detuning_mask")
        .constant(1.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(mask_assignments, use_hyperfine=True).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    detuning_op = np.array([0, 0, -1], dtype=int)
    detuning = sum(
        float(mask) * get_manybody_op(i, L, detuning_op)
        for i, mask in enumerate(detuning_mask)
    )

    assert np.all(hamiltonian.detuning_ops[0].get_diagonal(0.0) == detuning)

    emu_prog = EmulatorProgramCodeGen(
        mask_assignments, blockade_radius=6.2, use_hyperfine=True
    ).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    detuning_op_proj = project_to_subspace(detuning, hamiltonian.space.configurations)

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning_op_proj)

    # check hyperfine detuning value

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .hyperfine.detuning.scale("detuning_mask")
        .constant(1.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(mask_assignments).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    detuning_op = np.array([0, -1, 0], dtype=int)
    detuning = sum(
        float(mask) * get_manybody_op(i, L, detuning_op)
        for i, mask in enumerate(detuning_mask)
    )

    assert np.all(hamiltonian.detuning_ops[0].get_diagonal(0.0) == detuning)

    emu_prog = EmulatorProgramCodeGen(mask_assignments, blockade_radius=6.2).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    detuning_op_proj = project_to_subspace(detuning, hamiltonian.space.configurations)

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning_op_proj)


@pytest.mark.parametrize("L", L_VALUES)
def test_3_level_mask_rabi_real(L):
    rabi_mask = [Decimal(str(np.random.normal(0.0, 1.0))) for _ in range(L)]
    mask_assignments = dict(rabi_mask=rabi_mask)

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.scale("rabi_mask")
        .constant(1.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(mask_assignments, use_hyperfine=True).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0]], dtype=int)

    if L == 1:
        rabi = get_manybody_op(0, L, rabi_op)
    else:
        rabi = sum(
            float(mask) * get_manybody_op(i, L, rabi_op)
            for i, mask in enumerate(rabi_mask)
        )

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(
        mask_assignments, blockade_radius=6.2, use_hyperfine=True
    ).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)

    # check hyperfine rabi value

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .hyperfine.rabi.amplitude.scale("rabi_mask")
        .constant(1.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(mask_assignments).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]], dtype=int)

    if L == 1:
        rabi = get_manybody_op(0, L, rabi_op)
    else:
        rabi = sum(
            float(mask) * get_manybody_op(i, L, rabi_op)
            for i, mask in enumerate(rabi_mask)
        )

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(mask_assignments, blockade_radius=6.2).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


@pytest.mark.parametrize("L", L_VALUES)
def test_3_level_mask_complex(L):
    rabi_mask = [Decimal(str(np.random.normal(0.0, 1.0))) for _ in range(L)]
    mask_assignments = dict(
        rabi_mask=rabi_mask,
    )

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.scale("rabi_mask")
        .constant(1.0, 1.0)
        .phase.uniform.constant(0.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(mask_assignments, use_hyperfine=True).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0, 0], [0, 0, 0], [0, 1, 0]], dtype=int)

    if L == 1:
        rabi = get_manybody_op(0, L, rabi_op)
    else:
        rabi = sum(
            float(mask) * get_manybody_op(i, L, rabi_op)
            for i, mask in enumerate(rabi_mask)
        )

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(
        mask_assignments, blockade_radius=6.2, use_hyperfine=True
    ).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)

    # check hyperfine rabi value

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .hyperfine.rabi.amplitude.scale("rabi_mask")
        .constant(1.0, 1.0)
        .phase.uniform.constant(0.0, 1.0)
    ).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(mask_assignments).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0, 0], [1, 0, 0], [0, 0, 0]], dtype=int)

    if L == 1:
        rabi = get_manybody_op(0, L, rabi_op)
    else:
        rabi = sum(
            float(mask) * get_manybody_op(i, L, rabi_op)
            for i, mask in enumerate(rabi_mask)
        )

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(mask_assignments, blockade_radius=6.2).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


####################
# Test multi atom  #
####################


@pytest.mark.parametrize(
    ("sites", "L"),
    [
        (sites, L)
        for L in L_VALUES
        for n_sites in range(2, L + 1)
        for sites in combinations(range(L), n_sites)
    ],
)
def test_2_level_multi_atom_detuning(sites, L):
    program = Chain(L, lattice_spacing=6.1).rydberg.detuning
    detuning_coeffs = [Decimal(str(np.random.normal(0.0, 1.0))) for _ in sites]

    program = program.location(list(sites), detuning_coeffs)

    circuit = program.constant(1.0, 1.0).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    detuning_op = np.array([0, -1], dtype=int)

    detuning = sum(
        float(mask) * get_manybody_op(i, L, detuning_op)
        for i, mask in zip(sites, detuning_coeffs)
    )

    assert np.all(hamiltonian.detuning_ops[0].get_diagonal(0.0) == detuning)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    detuning_op_proj = project_to_subspace(detuning, hamiltonian.space.configurations)

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning_op_proj)


@pytest.mark.parametrize(
    ("sites", "L"),
    [
        (sites, L)
        for L in L_VALUES
        for n_sites in range(2, L + 1)
        for sites in combinations(range(L), n_sites)
    ],
)
def test_2_level_multi_atom_rabi_real(sites, L):
    program = Chain(L, lattice_spacing=6.1).rydberg.rabi.amplitude
    rabi_coeffs = [Decimal(str(np.random.normal(0.0, 1.0))) for _ in sites]

    program = program.location(list(sites), rabi_coeffs)

    circuit = program.constant(1.0, 1.0).parse_circuit()

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 1], [1, 0]], dtype=int)

    rabi = sum(
        float(mask) * get_manybody_op(i, L, rabi_op)
        for i, mask in zip(sites, rabi_coeffs)
    )

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


@pytest.mark.parametrize(
    ("sites", "L"),
    [
        (sites, L)
        for L in L_VALUES
        for n_sites in range(2, L + 1)
        for sites in combinations(range(L), n_sites)
    ],
)
def test_2_level_multi_atom_rabi_complex(sites, L):
    rabi_coeffs = [Decimal(str(np.random.normal(0.0, 1.0))) for _ in sites]

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.location(list(sites), rabi_coeffs)
        .constant(1.0, 1.0)
        .phase.uniform.constant(0.0, 1.0)
        .parse_circuit()
    )

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0], [1, 0]], dtype=int)

    rabi = sum(
        float(mask) * get_manybody_op(i, L, rabi_op)
        for i, mask in zip(sites, rabi_coeffs)
    )

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


@pytest.mark.parametrize(
    ("sites", "L"),
    [
        (sites, L)
        for L in L_VALUES
        for n_sites in range(2, L + 1)
        for sites in combinations(range(L), n_sites)
    ],
)
def test_3_level_multi_atom_detuning(sites, L):
    detuning_coeffs = [Decimal(str(np.random.normal(0.0, 1.0))) for site in sites]

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.detuning.location(list(sites), detuning_coeffs)
        .constant(1.0, 1.0)
        .parse_circuit()
    )

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(use_hyperfine=True).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    detuning_op = np.array([0, 0, -1], dtype=int)

    detuning = sum(
        float(mask) * get_manybody_op(i, L, detuning_op)
        for i, mask in zip(sites, detuning_coeffs)
    )

    assert np.all(hamiltonian.detuning_ops[0].get_diagonal(0.0) == detuning)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2, use_hyperfine=True).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    detuning_op_proj = project_to_subspace(detuning, hamiltonian.space.configurations)

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning_op_proj)

    # check hyperfine detuning value

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .hyperfine.detuning.location(list(sites), detuning_coeffs)
        .constant(1.0, 1.0)
        .parse_circuit()
    )

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    detuning_op = np.array([0, -1, 0], dtype=int)

    detuning = sum(
        float(mask) * get_manybody_op(i, L, detuning_op)
        for i, mask in zip(sites, detuning_coeffs)
    )

    assert np.all(hamiltonian.detuning_ops[0].get_diagonal(0.0) == detuning)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    detuning_op_proj = project_to_subspace(detuning, hamiltonian.space.configurations)

    assert np.all(hamiltonian.detuning_ops[0].diagonal == detuning_op_proj)


@pytest.mark.parametrize(
    ("sites", "L"),
    [
        (sites, L)
        for L in L_VALUES
        for n_sites in range(2, L + 1)
        for sites in combinations(range(L), n_sites)
    ],
)
def test_3_level_multi_atom_real(sites, L):
    rabi_coeffs = [Decimal(str(np.random.normal(0.0, 1.0))) for site in sites]

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.location(list(sites), rabi_coeffs)
        .constant(1.0, 1.0)
        .parse_circuit()
    )

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(use_hyperfine=True).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0]], dtype=int)

    rabi = sum(
        float(mask) * get_manybody_op(i, L, rabi_op)
        for i, mask in zip(sites, rabi_coeffs)
    )

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2, use_hyperfine=True).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)

    # check hyperfine rabi value

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .hyperfine.rabi.amplitude.location(list(sites), rabi_coeffs)
        .constant(1.0, 1.0)
        .parse_circuit()
    )

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]], dtype=int)

    rabi = sum(
        float(mask) * get_manybody_op(i, L, rabi_op)
        for i, mask in zip(sites, rabi_coeffs)
    )

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)


@pytest.mark.parametrize(
    ("sites", "L"),
    [
        (sites, L)
        for L in L_VALUES
        for n_sites in range(2, L + 1)
        for sites in combinations(range(L), n_sites)
    ],
)
def test_3_level_multi_atom_complex(sites, L):
    rabi_coeffs = [Decimal(str(np.random.normal(0.0, 1.0))) for site in sites]

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .rydberg.rabi.amplitude.location(list(sites), rabi_coeffs)
        .constant(1.0, 1.0)
        .phase.uniform.constant(0.0, 1.0)
        .parse_circuit()
    )

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen(use_hyperfine=True).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0, 0], [0, 0, 0], [0, 1, 0]], dtype=int)

    rabi = sum(
        float(mask) * get_manybody_op(i, L, rabi_op)
        for i, mask in zip(sites, rabi_coeffs)
    )

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2, use_hyperfine=True).emit(
        circuit
    )
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)

    # check hyperfine rabi value

    circuit = (
        Chain(L, lattice_spacing=6.1)
        .hyperfine.rabi.amplitude.location(list(sites), rabi_coeffs)
        .constant(1.0, 1.0)
        .phase.uniform.constant(0.0, 1.0)
        .parse_circuit()
    )

    cache = CompileCache()
    emu_prog = EmulatorProgramCodeGen().emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen(cache).emit(emu_prog)

    rabi_op = np.array([[0, 0, 0], [1, 0, 0], [0, 0, 0]], dtype=int)

    rabi = sum(
        float(mask) * get_manybody_op(i, L, rabi_op)
        for i, mask in zip(sites, rabi_coeffs)
    )

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi)

    emu_prog = EmulatorProgramCodeGen(blockade_radius=6.2).emit(circuit)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emu_prog)

    rabi_op_proj = project_to_subspace(rabi, hamiltonian.space.configurations)

    assert np.all(hamiltonian.rabi_ops[0].op.tocsr().toarray() == rabi_op_proj)



import simplejson as json
import os
from bloqade.submission.ir.capabilities import QuEraCapabilities


# TODO: Create unit converter for capabilities
def get_capabilities() -> QuEraCapabilities:
    base_path = os.path.dirname(__file__)
    full_path = os.path.join(
        base_path, "quera_api_client", "config", "capabilities.json"
    )
    with open(full_path, "r") as io:
        return QuEraCapabilities(**json.load(io))



import os
import json


def load_config(qpu: str):
    real_path = os.path.realpath(__file__)
    real_path_list = os.path.split(real_path)[:-1]
    real_path = os.path.join(*real_path_list)

    if qpu == "Aquila":
        with open(
            os.path.join(real_path, "config", "aquila_api_config.json"), "r"
        ) as f:
            return json.load(f)
    elif qpu == "Mock":
        with open(os.path.join(real_path, "config", "mock_api_config.json"), "r") as f:
            return json.load(f)
    else:
        raise NotImplementedError(
            f"QPU {qpu} is not supported. Supported QPUs are Aquila and Mock."
        )



from typing import Any, Optional
import numpy as np
from pydantic.dataclasses import dataclass
from pydantic import ValidationError, validator
from .tree_print import Printer
import re
from decimal import Decimal
import numbers

__all__ = [
    "var",
    "cast",
    "Scalar",
    "Interval",
    "Variable",
    "Literal",
]


@dataclass(frozen=True)
class Scalar:
    """Base class for all scalar expressions.

    ```bnf
    <scalar> ::= <literal>
    | <variable>
    | <default>
    | <negative>
    | <add>
    | <mul>
    | <min>
    | <max>
    | <slice>
    | <inverval>

    <mul> ::= <scalar> '*' <scalar>
    <add> ::= <scalar> '+' <scalar>
    <min> ::= 'min' <scalar>+
    <max> ::= 'max' <scalar>+
    <slice> ::= <scalar expr> '[' <interval> ']'
    <interval> ::= <scalar expr> '..' <scalar expr>
    <real> ::= <literal> | <var>
    ```
    """

    def __getitem__(self, s: slice) -> "Scalar":
        return Scalar.canonicalize(Slice(self, Interval.from_slice(s)))

    def __add__(self, other: "Scalar") -> "Scalar":
        return self.add(other)

    def __sub__(self, other: "Scalar") -> "Scalar":
        return self.sub(other)

    def __mul__(self, other: "Scalar") -> "Scalar":
        return self.mul(other)

    def __truediv__(self, other: "Scalar") -> "Scalar":
        return self.div(other)

    def __radd__(self, other: "Scalar") -> "Scalar":
        try:
            lhs = cast(other)
        except TypeError:
            return NotImplemented

        if not isinstance(lhs, Scalar):
            return NotImplemented

        return lhs.add(self)

    def __rsub__(self, other: "Scalar") -> "Scalar":
        try:
            lhs = cast(other)
        except TypeError:
            return NotImplemented

        if not isinstance(lhs, Scalar):
            return NotImplemented

        return lhs.sub(self)

    def __rmul__(self, other: "Scalar") -> "Scalar":
        try:
            lhs = cast(other)
        except TypeError:
            return NotImplemented

        if not isinstance(lhs, Scalar):
            return NotImplemented

        return lhs.mul(self)

    def __rtruediv__(self, other: "Scalar") -> "Scalar":
        try:
            lhs = cast(other)
        except TypeError:
            return NotImplemented

        if not isinstance(lhs, Scalar):
            return NotImplemented

        return lhs.div(self)

    def __neg__(self) -> "Scalar":
        return Scalar.canonicalize(Negative(self))

    def add(self, other) -> "Scalar":
        try:
            rhs = cast(other)
        except TypeError:
            return NotImplemented

        if not isinstance(rhs, Scalar):
            return NotImplemented

        expr = Add(lhs=self, rhs=rhs)
        return Scalar.canonicalize(expr)

    def sub(self, other) -> "Scalar":
        try:
            rhs = cast(other)
        except TypeError:
            return NotImplemented

        if not isinstance(rhs, Scalar):
            return NotImplemented

        expr = Add(lhs=self, rhs=-rhs)
        return Scalar.canonicalize(expr)

    def mul(self, other) -> "Scalar":
        try:
            rhs = cast(other)
        except TypeError:
            return NotImplemented

        if not isinstance(rhs, Scalar):
            return NotImplemented

        expr = Mul(lhs=self, rhs=rhs)
        return Scalar.canonicalize(expr)

    def div(self, other) -> "Scalar":
        try:
            rhs = cast(other)
        except TypeError:
            return NotImplemented

        if not isinstance(rhs, Scalar):
            return NotImplemented

        expr = Div(lhs=self, rhs=rhs)
        return Scalar.canonicalize(expr)

    def min(self, other) -> "Scalar":
        try:
            other_expr = cast(other)
        except TypeError:
            return NotImplemented

        expr = Min(exprs=frozenset({self, other_expr}))
        return Scalar.canonicalize(expr)

    def max(self, other) -> "Scalar":
        try:
            other_expr = cast(other)
        except TypeError:
            return NotImplemented

        expr = Max(exprs=frozenset({self, other_expr}))
        return Scalar.canonicalize(expr)

    def __str__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    @staticmethod
    def canonicalize(expr: "Scalar") -> "Scalar":
        def minmax(op, exprs):
            new_exprs = set()
            for expr in exprs:
                if isinstance(expr, op):
                    exprs = list(map(Scalar.canonicalize, exprs))
                    new_exprs.update(exprs)
                else:
                    expr = Scalar.canonicalize(expr)
                    new_exprs.add(expr)

            if len(new_exprs) > 1:
                return op(exprs=frozenset(new_exprs))
            else:
                (new_expr,) = new_exprs
                return new_expr

        if isinstance(expr, Negative):
            sub_expr = expr.expr
            if isinstance(sub_expr, Negative):
                return Scalar.canonicalize(sub_expr.expr)
            elif isinstance(sub_expr, Literal) and sub_expr.value < 0:
                return Literal(-sub_expr.value)
        elif isinstance(expr, Add):
            lhs = expr.lhs
            rhs = expr.rhs
            if isinstance(lhs, Literal) and lhs.value == 0:
                return Scalar.canonicalize(rhs)
            elif isinstance(rhs, Literal) and rhs.value == 0:
                return Scalar.canonicalize(lhs)
            elif isinstance(lhs, Literal) and isinstance(rhs, Literal):
                return Literal(lhs.value + rhs.value)
            elif (
                isinstance(lhs, Negative)
                and isinstance(lhs.expr, Literal)
                and isinstance(rhs, Literal)
            ):
                return Literal(rhs.value - lhs.expr.value)
            elif (
                isinstance(rhs, Negative)
                and isinstance(rhs.expr, Literal)
                and isinstance(lhs, Literal)
            ):
                return Literal(lhs.value - rhs.expr.value)
            elif isinstance(lhs, Negative) and lhs.expr == rhs:
                return Literal(0.0)
            elif isinstance(rhs, Negative) and rhs.expr == lhs:
                return Literal(0.0)
        elif isinstance(expr, Mul):
            lhs = expr.lhs
            rhs = expr.rhs
            if isinstance(lhs, Literal) and lhs.value == 1:
                return Scalar.canonicalize(rhs)
            elif isinstance(rhs, Literal) and rhs.value == 1:
                return Scalar.canonicalize(lhs)
            elif isinstance(lhs, Literal) and isinstance(rhs, Literal):
                return Literal(lhs.value * rhs.value)
            elif isinstance(lhs, Literal) and lhs.value == 0:
                return Literal(0.0)
            elif isinstance(rhs, Literal) and rhs.value == 0:
                return Literal(0.0)
        elif isinstance(expr, Div):
            lhs = expr.lhs
            rhs = expr.rhs
            if isinstance(lhs, Literal) and lhs.value == 0:
                return Literal(0.0)
            elif isinstance(rhs, Literal) and rhs.value == 1:
                return Scalar.canonicalize(lhs)
            elif isinstance(lhs, Literal) and isinstance(rhs, Literal):
                return Literal(lhs.value / rhs.value)
        elif isinstance(expr, Min):
            return minmax(Min, expr.exprs)
        elif isinstance(expr, Max):
            return minmax(Max, expr.exprs)

        return expr


def check_variable_name(name: str) -> None:
    regex = "^[A-Za-z_][A-Za-z0-9_]*"
    re_match = re.match(regex, name)
    if re_match.group() != name:
        raise ValidationError(f"string '{name}' is not a valid python identifier")


def cast(py) -> "Scalar":
    """
    1. cast Real number (or list/tuple of Real numbers)
    to [`Scalar Literal`][bloqade.ir.scalar.Literal].

    2. cast str (or list/tuple of Real numbers)
    to [`Scalar Variable`][bloqade.ir.scalar.Variable].

    Args:
        py (Union[str,Real,Tuple[Real],List[Real]]): python object to cast

    Returns:
        Scalar
    """
    ret = trycast(py)
    if ret is None:
        raise TypeError(f"Cannot cast {type(py)} to Scalar Literal")

    return ret


# TODO: RL: remove support on List and Tuple just use map?
#       this is making type inference much harder to parse
#       in human brain
# [KHW] it need to be there. For recursive replace for nested
#       list/tuple
def trycast(py) -> Optional[Scalar]:
    if isinstance(py, (int, bool, numbers.Real)):
        return Literal(Decimal(str(py)))
    elif isinstance(py, Decimal):
        return Literal(py)
    elif isinstance(py, str):
        return Variable(py)
    elif isinstance(py, Scalar):
        return py
    elif isinstance(py, (list, tuple)):
        return type(py)(map(cast, py))
    elif isinstance(py, np.ndarray):
        return np.array(list(map(cast, py)))
    else:
        return


def var(py: str) -> "Variable":
    """cast string (or list/tuple of strings)
    to [`Variable`][bloqade.ir.scalar.Variable].

    Args:
        py (Union[str, List[str]]): a string or list/tuple of strings

    Returns:
       Union[Variable]
    """
    ret = tryvar(py)
    if ret is None:
        raise TypeError(f"Cannot cast {type(py)} to Variable")

    return ret


def tryvar(py) -> Optional["Variable"]:
    if isinstance(py, str):
        return Variable(py)
    if isinstance(py, Variable):
        return py
    elif isinstance(py, (list, tuple)):
        return type(py)(map(var, py))
    else:
        return


class Real(Scalar):
    # """Base class for all real expressions."""
    pass


@dataclass(frozen=True)
class Literal(Real):
    value: Decimal
    """Scalar Literal, which stores a decimaal value instance.

    Args:
        value (Decimal): decimal value instance

    """

    def __call__(self, **assignments) -> Decimal:
        return self.value

    def __str__(self) -> str:
        return str(self.value)

    def children(self):
        return []

    def print_node(self):
        return f"Literal: {self.value}"

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)


@dataclass(frozen=True)
class Variable(Real):
    """Variable, which stores a variable name.

    Args:
        name (str): variable instance.

    """

    name: str

    def __call__(self, **assignments) -> Decimal:
        if self.name not in assignments:
            raise ValueError(f"Variable {self.name} not assigned")

        return Decimal(str(assignments[self.name]))

    def __str__(self) -> str:
        return self.name

    def children(self):
        return []

    def print_node(self):
        return f"Variable: {self.name}"

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    @validator("name", allow_reuse=True)
    def validate_name(cls, name):
        check_variable_name(name)
        if name in ["__batch_params"]:
            raise ValidationError(
                "Cannot use reserved name `__batch_params` for variable name"
            )

        return name


@dataclass(frozen=True)
class AssignedVariable(Scalar):
    name: str
    value: Decimal

    def __call__(self, **assignments) -> Decimal:
        if self.name in assignments:
            raise ValueError(f"Variable {self.name} already assigned")

        return self.value

    def __str__(self) -> str:
        return f"{self.name}"

    def children(self):
        return []

    def print_node(self):
        return f"AssignedVariable: {self.name} = {self.value!s}"

    @validator("name", allow_reuse=True)
    def validate_name(cls, name):
        check_variable_name(name)
        if name in ["__batch_params"]:
            raise ValidationError(
                "Cannot use reserved name `__batch_params` for variable name"
            )

        return name


@dataclass(frozen=True)
class Negative(Scalar):
    expr: Scalar

    def __call__(self, **assignments) -> Decimal:
        return -self.expr(**assignments)

    def __str__(self) -> str:
        return f"-({self.expr!s})"

    def children(self):
        return [self.expr]

    def print_node(self):
        return "Negative"


@dataclass(frozen=True)
class Interval:
    start: Optional[Scalar]
    stop: Optional[Scalar]

    @staticmethod
    def from_slice(s: slice) -> "Interval":
        start, stop, step = s.start, s.stop, s.step

        if start is None and stop is None and step is None:
            raise ValueError("Slice must have at least one argument")
        elif step is not None:
            raise ValueError("Slice step must be None")

        else:
            if start is None:
                start = None
            else:
                start = cast(start)
                if not isinstance(start, Scalar):
                    raise ValueError("Slice start must be Scalar")

            if stop is None:
                stop = None
            else:
                stop = cast(stop)
                if not isinstance(stop, Scalar):
                    raise ValueError("Slice stop must be Scalar")

            return Interval(start, stop)

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def __str__(self):
        if self.start is None:
            if self.stop is None:
                raise ValueError("Interval must have at least one bound")
            else:
                return f":{str(self.stop)}"
        else:
            if self.stop is None:
                return f"{str(self.start)}:"
            else:
                return f"{str(self.start)}:{str(self.stop)}"

    def print_node(self):
        return "Interval"

    def children(self):
        if self.start is None:
            if self.stop is None:
                raise ValueError("Interval must have at least one bound")
            else:
                return {"stop": self.stop}
        else:
            if self.stop is None:
                return {"start": self.start}
            else:
                return {"start": self.start, "stop": self.stop}


@dataclass(frozen=True)
class Slice(Scalar):
    expr: Scalar  # duration
    interval: Interval

    def __call__(self, **assignments) -> Decimal:
        dur = self.expr(**assignments)
        start = (
            self.interval.start(**assignments)
            if self.interval.start is not None
            else Decimal("0")
        )
        stop = (
            self.interval.stop(**assignments) if self.interval.stop is not None else dur
        )

        if start < 0:
            raise ValueError(
                f"Slice start must be non-negative, got {start} from expr:\n"
                f"{repr(self.interval.start)}\n"
                f"with assignments: {assignments}"
            )

        if stop > dur:
            raise ValueError(
                "Slice stop must be smaller or equal to than duration "
                f"{dur}, got {stop} from expr:\b"
                f"{repr(self.interval.stop)}\n"
                f"with assignments: {assignments}"
            )

        ret = stop - start

        if ret < 0:
            raise ValueError(
                f"start is larger than stop, get start = {start} and stop = {stop}\n"
                "from start expr:\n"
                f"{repr(self.interval.start)}\n"
                "and stop expr:\n"
                f"{repr(self.interval.stop)}\n"
                f"with assignments: {assignments}"
            )

        return ret

    def __str__(self) -> str:
        return f"({self.expr!s})[{self.interval!s}]"

    def children(self):
        return {"Scalar": self.expr, None: self.interval}

    def print_node(self):
        return "Slice"


@dataclass(frozen=True)
class Add(Scalar):
    lhs: Scalar
    rhs: Scalar

    def __call__(self, **assignments) -> Decimal:
        return self.lhs(**assignments) + self.rhs(**assignments)

    def __str__(self) -> str:
        return f"({self.lhs!s} + {self.rhs!s})"

    def children(self):
        return [self.lhs, self.rhs]

    def print_node(self):
        return "+"


@dataclass(frozen=True)
class Mul(Scalar):
    lhs: Scalar
    rhs: Scalar

    def __call__(self, **assignments) -> Decimal:
        return self.lhs(**assignments) * self.rhs(**assignments)

    def __str__(self) -> str:
        return f"({self.lhs!s} * {self.rhs!s})"

    def children(self):
        return [self.lhs, self.rhs]

    def print_node(self):
        return "*"


@dataclass(frozen=True)
class Div(Scalar):
    lhs: Scalar
    rhs: Scalar

    def __call__(self, **assignments) -> Decimal:
        return self.lhs(**assignments) / self.rhs(**assignments)

    def __str__(self) -> str:
        return f"({self.lhs!s} / {self.rhs!s})"

    def children(self):
        return [self.lhs, self.rhs]

    def print_node(self):
        return "/"


@dataclass(frozen=True)
class Min(Scalar):
    exprs: frozenset[Scalar]

    def __call__(self, **assignments) -> Any:
        return min(expr(**assignments) for expr in self.exprs)

    def __str__(self) -> str:
        return f"min({', '.join(map(str, self.exprs))})"

    def children(self):
        return list(self.exprs)

    def print_node(self):
        return "min"


@dataclass(frozen=True)
class Max(Scalar):
    exprs: frozenset[Scalar]

    def __call__(self, **assignments) -> Any:
        return max(expr(**assignments) for expr in self.exprs)

    def children(self):
        return list(self.exprs)

    def print_node(self):
        return "max"

    def __str__(self) -> str:
        return f"max({', '.join(map(str, self.exprs))})"



from dataclasses import dataclass
from functools import cached_property
from scipy.sparse import csr_matrix, csc_matrix, coo_matrix
import numpy as np
from numpy.typing import NDArray
from beartype.typing import Union
from numba import njit

# from beartype.vale import IsAttr, IsEqual
# from beartype import beartype
# from typing import Annotated


@njit(cache=True)
def _csc_matvec_impl(ncol, data, indices, indptr, scale, input, output):
    for i in range(ncol):
        for j in range(indptr[i], indptr[i + 1]):
            output[indices[j]] += scale * data[j] * input[i]

    return output


@njit(cache=True)
def _csr_matvec_impl(nrow, data, indices, indptr, scale, input, output):
    zero = type(output[0])(0)
    for i in range(nrow):
        row_out = zero
        for j in range(indptr[i], indptr[i + 1]):
            row_out += data[j] * input[indices[j]]

        output[i] += scale * row_out

    return output


@dataclass(frozen=True)
class SparseMatrixCSC:
    data: NDArray
    indices: NDArray
    indptr: NDArray
    shape: tuple[int, int]

    @staticmethod
    def create(arg1, shape=None, dtype=None, copy=False) -> "SparseMatrixCSC":
        mat = csc_matrix(arg1, shape=shape, dtype=dtype, copy=copy)
        return SparseMatrixCSC(mat.data, mat.indices, mat.indptr, mat.shape)

    def tocsr(self):
        return csc_matrix((self.data, self.indices, self.indptr), self.shape).tocsr()

    def toarray(self):
        return self.tocsr().toarray()

    @property
    def T(self):
        return SparseMatrixCSR(self.data, self.indices, self.indptr, self.shape[::-1])

    def matvec(self, other, out=None, scale=1):
        if out is None:
            out = np.zeros_like(other, dtype=np.result_type(scale, self.data, other))

        return _csc_matvec_impl(
            self.shape[1], self.data, self.indices, self.indptr, scale, other, out
        )


@dataclass(frozen=True)
class SparseMatrixCSR:
    data: NDArray
    indices: NDArray
    indptr: NDArray
    shape: tuple[int, int]

    @staticmethod
    def create(arg1, shape=None, dtype=None, copy=False) -> "SparseMatrixCSR":
        mat = csr_matrix(arg1, shape=shape, dtype=dtype, copy=copy)
        return SparseMatrixCSR(mat.data, mat.indices, mat.indptr, mat.shape)

    def tocsr(self):
        return csr_matrix((self.data, self.indices, self.indptr), self.shape)

    def toarray(self):
        return self.tocsr().toarray()

    @property
    def T(self):
        return SparseMatrixCSC(self.data, self.indices, self.indptr, self.shape[::-1])

    def matvec(self, other, out=None, scale=1):
        if out is None:
            out = np.zeros_like(other, dtype=np.result_type(scale, self.data, other))

        return _csr_matvec_impl(
            self.shape[0], self.data, self.indices, self.indptr, scale, other, out
        )


@njit(cache=True)
def _index_mapping_row_col_sliced(scale, input, output):
    for i in range(output.size):
        output[i] += scale * input[i]

    return output


@njit(cache=True)
def _index_mapping_row_sliced(col_indices, scale, input, output):
    for i in range(col_indices.size):
        output[i] += scale * input[col_indices[i]]

    return output


@njit(cache=True)
def _index_mapping_col_sliced(row_indices, scale, input, output):
    for i in range(row_indices.size):
        output[row_indices[i]] += scale * input[i]

    return output


@njit(cache=True)
def _index_mapping_impl(col_indices, row_indices, scale, input, output):
    for i in range(row_indices.size):
        output[row_indices[i]] += scale * input[col_indices[i]]

    return output


# use csr_matrix/csc_matrix for rabi-terms that span multiple sites.
# use IndexMapping for local rabi-terms
@dataclass(frozen=True)
class IndexMapping:
    n_row: int
    row_indices: Union[NDArray, slice]
    col_indices: Union[NDArray, slice]

    @property
    def T(self) -> "IndexMapping":
        return IndexMapping(self.n_row, self.col_indices, self.row_indices)

    @cached_property
    def _matvec_dispatcher(self):
        if isinstance(self.col_indices, slice) and isinstance(self.row_indices, slice):

            def _matvec_imp(col_indices, row_indices, scale, input, output):
                return _index_mapping_row_col_sliced(
                    scale, input[col_indices], output[row_indices]
                )

        elif isinstance(self.col_indices, slice):

            def _matvec_imp(col_indices, row_indices, scale, input, output):
                return _index_mapping_col_sliced(
                    row_indices, scale, input[col_indices], output
                )

        elif isinstance(self.row_indices, slice):

            def _matvec_imp(col_indices, row_indices, scale, input, output):
                return _index_mapping_row_sliced(
                    col_indices, scale, input, output[row_indices]
                )

        else:

            def _matvec_imp(col_indices, row_indices, scale, input, output):
                return _index_mapping_impl(
                    col_indices, row_indices, scale, input, output
                )

        return _matvec_imp

    def matvec(self, other, out=None, scale=1):
        if out is None:
            out = np.zeros_like(other, dtype=np.result_type(scale, other))

        return self._matvec_dispatcher(
            self.col_indices, self.row_indices, scale, other, out
        )

    def tocoo(self) -> coo_matrix:
        if isinstance(self.row_indices, slice):
            row_slice = self.row_indices
            start = row_slice.start if row_slice.start is not None else 0
            stop = row_slice.stop if row_slice.stop is not None else self.n_row
            step = row_slice.step if row_slice.step is not None else 1
            rows = np.arange(start, stop, step)
        elif (
            isinstance(self.row_indices, np.ndarray) and self.row_indices.dtype == bool
        ):
            rows = np.where(self.row_indices)[0]
        else:
            rows = self.row_indices

        if isinstance(self.col_indices, slice):
            col_slice = self.col_indices
            start = col_slice.start if col_slice.start is not None else 0
            stop = col_slice.stop if col_slice.stop is not None else self.n_row
            step = col_slice.step if col_slice.step is not None else 1

            cols = np.arange(start, stop, step)
        elif (
            isinstance(self.col_indices, np.ndarray) and self.col_indices.dtype == bool
        ):
            cols = np.where(self.col_indices)[0]
        else:
            cols = self.col_indices

        data = np.ones(rows.size)

        return coo_matrix((data, (rows, cols)), shape=(self.n_row, self.n_row))

    def tocsc(self):
        return self.tocoo().tocsc()

    def tocsr(self):
        return self.tocoo().tocsr()



from bloqade.ir.analysis.assignment_scan import AssignmentScan
from bloqade.ir.location.base import AtomArrangement, SiteFilling
from bloqade.ir.visitor.analog_circuit import AnalogCircuitVisitor
from bloqade.ir.visitor.waveform import WaveformVisitor
from bloqade.ir.control.field import (
    AssignedRunTimeVector,
    Field,
    ScaledLocations,
    RunTimeVector,
    UniformModulation,
)
import bloqade.ir.control.sequence as sequence
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.waveform as waveform
import bloqade.ir.control.field as field
import bloqade.ir as ir
from bloqade.ir.analysis.is_hyperfine import IsHyperfineSequence
from bloqade.emulate.ir.atom_type import ThreeLevelAtom, TwoLevelAtom
from bloqade.emulate.ir.emulator import (
    DetuningOperatorData,
    LevelCoupling,
    RabiOperatorData,
    RabiOperatorType,
    DetuningTerm,
    CompiledWaveform,
    EmulatorProgram,
    Register,
    Fields,
    RabiTerm,
)

from typing import Any, Dict, Optional, Union, List
from numbers import Number, Real
from decimal import Decimal

ParamType = Union[Real, List[Real]]


class WaveformCompiler(WaveformVisitor):
    # TODO: implement AST generator for waveforms.
    def __init__(self, assignments: Dict[str, Number] = {}):
        self.assignments = assignments

    def emit(self, ast: waveform.Waveform) -> CompiledWaveform:
        return CompiledWaveform(self.assignments, ast)


class EmulatorProgramCodeGen(AnalogCircuitVisitor):
    def __init__(
        self,
        assignments: Dict[str, Number] = {},
        blockade_radius: Real = 0.0,
        use_hyperfine: bool = False,
    ):
        self.blockade_radius = Decimal(str(blockade_radius))
        self.assignments = assignments
        self.waveform_compiler = WaveformCompiler(assignments)
        self.register = None
        self.duration = 0.0
        self.pulses = {}
        self.level_couplings = set()
        self.original_index = []
        self.is_hyperfine = use_hyperfine

    def visit_analog_circuit(self, ast: ir.AnalogCircuit):
        self.visit(ast.register)
        self.visit(ast.sequence)

    def visit_register(self, ast: AtomArrangement) -> Any:
        positions = []
        for org_index, loc_info in enumerate(ast.enumerate()):
            if loc_info.filling == SiteFilling.filled:
                position = tuple([pos(**self.assignments) for pos in loc_info.position])
                positions.append(position)
                self.original_index.append(org_index)

        if self.is_hyperfine:
            self.register = Register(
                ThreeLevelAtom,
                positions,
                blockade_radius=self.blockade_radius,
            )
        else:
            self.register = Register(
                TwoLevelAtom,
                positions,
                blockade_radius=self.blockade_radius,
            )

    def visit_sequence(self, ast: sequence.Sequence) -> None:
        level_coupling_mapping = {
            sequence.hyperfine: LevelCoupling.Hyperfine,
            sequence.rydberg: LevelCoupling.Rydberg,
        }
        for level_coupling, sub_pulse in ast.pulses.items():
            self.visit(sub_pulse)
            self.pulses[level_coupling_mapping[level_coupling]] = Fields(
                detuning=self.detuning_terms,
                rabi=self.rabi_terms,
            )

    def visit_named_sequence(self, ast: sequence.NamedSequence) -> None:
        self.vicit(ast.sequence)

    def visit_slice_sequence(self, ast: sequence.Slice) -> None:
        raise NotImplementedError("Slice sequences are not supported by the emulator.")

    def visit_append_sequence(self, ast: sequence.Append) -> None:
        raise NotImplementedError("Append sequences are not supported by the emulator.")

    def visit_pulse(self, ast: pulse.Pulse) -> None:
        detuning = ast.fields.get(pulse.detuning)
        amplitude = ast.fields.get(pulse.rabi.amplitude)
        phase = ast.fields.get(pulse.rabi.phase)

        self.detuning_terms = self.visit_detuning(detuning)
        self.rabi_terms = self.visit_rabi(amplitude, phase)

    def visit_named_pulse(self, ast: pulse.NamedPulse) -> Any:
        self.visit(ast.pulse)

    def visit_slice_pulse(self, ast: pulse.Slice) -> Any:
        raise NotImplementedError("Slice pulses are not supported by the emulator.")

    def visit_append_pulse(self, ast: pulse.Append) -> Any:
        raise NotImplementedError("Append pulses are not supported by the emulator.")

    def visit_uniform_modulation(self, ast: UniformModulation) -> Dict[int, Decimal]:
        return {atom: Decimal("1.0") for atom in range(self.n_atoms)}

    def visit_run_time_vector(self, ast: RunTimeVector) -> Dict[int, Decimal]:
        value = self.assignments[ast.name]
        return {
            new_index: Decimal(str(value[original_index]))
            for new_index, original_index in enumerate(self.original_index)
            if value[original_index] != 0
        }

    def visit_assigned_run_time_vector(
        self, ast: AssignedRunTimeVector
    ) -> Dict[int, Decimal]:
        return {
            new_index: Decimal(str(ast.value[original_index]))
            for new_index, original_index in enumerate(self.original_index)
            if ast.value[original_index] != 0
        }

    def visit_scaled_locations(self, ast: ScaledLocations) -> Dict[int, Decimal]:
        target_atoms = {}

        for location in ast.value.keys():
            if location.value >= self.n_sites or location.value < 0:
                raise ValueError(
                    f"Location {location.value} is out of bounds for register with "
                    f"{self.n_sites} sites."
                )

        for new_index, original_index in enumerate(self.original_index):
            value = ast.value.get(field.Location(original_index))
            if value is not None and value != 0:
                target_atoms[new_index] = value(**self.assignments)

        return target_atoms

    def visit_detuning(self, ast: Optional[Field]):
        if ast is None:
            return []

        terms = []

        if len(ast.drives) <= self.n_atoms:
            for sm, wf in ast.drives.items():
                self.duration = max(
                    float(wf.duration(**self.assignments)), self.duration
                )

                terms.append(
                    DetuningTerm(
                        operator_data=DetuningOperatorData(target_atoms=self.visit(sm)),
                        amplitude=self.waveform_compiler.emit(wf),
                    )
                )
        else:
            target_atom_dict = {sm: self.visit(sm) for sm in ast.drives.keys()}

            for atom in range(self.n_atoms):
                if not any(atom in value for value in target_atom_dict.values()):
                    continue

                wf = sum(
                    (
                        target_atom_dict[sm][atom] * wf
                        for sm, wf in ast.drives.items()
                        if atom in target_atom_dict[sm]
                    ),
                    start=waveform.Constant(0.0, 0.0),
                )

                self.duration = max(
                    float(wf.duration(**self.assignments)), self.duration
                )

                terms.append(
                    DetuningTerm(
                        operator_data=DetuningOperatorData(
                            target_atoms={atom: Decimal("1.0")}
                        ),
                        amplitude=self.waveform_compiler.emit(wf),
                    )
                )

        return terms

    def visit_rabi(self, amplitude: Optional[Field], phase: Optional[Field]):
        terms = []

        if amplitude is None:
            return terms

        if phase is None and len(amplitude.drives) <= self.n_atoms:
            for sm, wf in amplitude.drives.items():
                self.duration = max(
                    float(wf.duration(**self.assignments)), self.duration
                )

                target_atoms = self.visit(sm)

                if len(target_atoms) == 0:
                    continue
                elif len(target_atoms) == 1:
                    (scale,) = target_atoms.values()
                    if scale != 1:
                        wf = scale * wf

                terms.append(
                    RabiTerm(
                        operator_data=RabiOperatorData(
                            target_atoms=target_atoms,
                            operator_type=RabiOperatorType.RabiSymmetric,
                        ),
                        amplitude=self.waveform_compiler.emit(wf),
                    )
                )
        elif phase is None:  # fully local real rabi fields
            amplitude_target_atoms_dict = {
                sm: self.visit(sm) for sm in amplitude.drives.keys()
            }
            for atom in range(self.n_atoms):
                if not any(
                    atom in value for value in amplitude_target_atoms_dict.values()
                ):
                    continue

                amplitude_wf = sum(
                    (
                        amplitude_target_atoms_dict[sm][atom] * wf
                        for sm, wf in amplitude.drives.items()
                        if atom in amplitude_target_atoms_dict[sm]
                    ),
                    start=waveform.Constant(0.0, 0.0),
                )

                self.duration = max(
                    float(amplitude_wf.duration(**self.assignments)), self.duration
                )

                terms.append(
                    RabiTerm(
                        operator_data=RabiOperatorData(
                            target_atoms={atom: Decimal("1.0")},
                            operator_type=RabiOperatorType.RabiSymmetric,
                        ),
                        amplitude=self.waveform_compiler.emit(amplitude_wf),
                    )
                )
        elif (
            len(phase.drives) == 1
            and UniformModulation() in phase.drives
            and len(amplitude.drives) <= self.n_atoms
        ):
            (phase_waveform,) = phase.drives.values()
            rabi_phase = self.waveform_compiler.emit(phase_waveform)
            self.duration = max(
                float(phase_waveform.duration(**self.assignments)), self.duration
            )
            for sm, wf in amplitude.drives.items():
                self.duration = max(
                    float(wf.duration(**self.assignments)), self.duration
                )

                target_atoms = self.visit(sm)

                if len(target_atoms) == 0:
                    continue
                elif len(target_atoms) == 1:
                    (scale,) = target_atoms.values()
                    if scale != 1:
                        wf = scale * wf

                terms.append(
                    RabiTerm(
                        operator_data=RabiOperatorData(
                            target_atoms=target_atoms,
                            operator_type=RabiOperatorType.RabiAsymmetric,
                        ),
                        amplitude=self.waveform_compiler.emit(wf),
                        phase=rabi_phase,
                    )
                )
        else:
            phase_target_atoms_dict = {sm: self.visit(sm) for sm in phase.drives.keys()}
            amplitude_target_atoms_dict = {
                sm: self.visit(sm) for sm in amplitude.drives.keys()
            }

            terms = []
            for atom in range(self.n_atoms):
                if not any(
                    atom in value for value in amplitude_target_atoms_dict.values()
                ):
                    continue

                phase_wf = sum(
                    (
                        phase_target_atoms_dict[sm][atom] * wf
                        for sm, wf in phase.drives.items()
                        if atom in phase_target_atoms_dict[sm]
                    ),
                    start=waveform.Constant(0.0, 0.0),
                )

                amplitude_wf = sum(
                    (
                        amplitude_target_atoms_dict[sm][atom] * wf
                        for sm, wf in amplitude.drives.items()
                        if atom in amplitude_target_atoms_dict[sm]
                    ),
                    start=waveform.Constant(0.0, 0.0),
                )

                self.duration = max(
                    float(amplitude_wf.duration(**self.assignments)), self.duration
                )
                self.duration = max(
                    float(phase_wf.duration(**self.assignments)), self.duration
                )

                terms.append(
                    RabiTerm(
                        operator_data=RabiOperatorData(
                            target_atoms={atom: 1},
                            operator_type=RabiOperatorType.RabiAsymmetric,
                        ),
                        amplitude=self.waveform_compiler.emit(amplitude_wf),
                        phase=self.waveform_compiler.emit(phase_wf),
                    )
                )

        return terms

    def emit(self, circuit: ir.AnalogCircuit) -> EmulatorProgram:
        self.assignments = AssignmentScan(self.assignments).emit(circuit.sequence)
        self.is_hyperfine = IsHyperfineSequence().emit(circuit) or self.is_hyperfine
        self.n_atoms = circuit.register.n_atoms
        self.n_sites = circuit.register.n_sites

        self.visit(circuit)
        return EmulatorProgram(
            register=self.register,
            duration=self.duration,
            pulses=self.pulses,
        )



from bloqade.constants import RB_C6
from bloqade.emulate.ir.emulator import (
    DetuningOperatorData,
    EmulatorProgram,
    LevelCoupling,
    Register,
    Fields,
    DetuningTerm,
    RabiOperatorData,
    RabiOperatorType,
    RabiTerm,
    Visitor,
)
from bloqade.emulate.ir.space import Space
from bloqade.emulate.ir.atom_type import (
    TwoLevelAtomType,
    ThreeLevelAtomType,
)
from bloqade.emulate.ir.state_vector import (
    RabiOperator,
    DetuningOperator,
    RydbergHamiltonian,
)
from bloqade.emulate.sparse_operator import (
    IndexMapping,
    SparseMatrixCSR,
)
from scipy.sparse import csr_matrix
import numpy as np
from numpy.typing import NDArray
from typing import Dict, Optional, Tuple, Union
from dataclasses import dataclass, field

OperatorData = Union[DetuningOperatorData, RabiOperatorData]
MatrixTypes = Union[csr_matrix, IndexMapping, NDArray]


@dataclass
class CompileCache:
    """This class is used to cache the results of the code generation."""

    operator_cache: Dict[Tuple[Register, OperatorData], MatrixTypes] = field(
        default_factory=dict
    )
    space_cache: Dict[Register, Tuple[Space, NDArray]] = field(default_factory=dict)


class RydbergHamiltonianCodeGen(Visitor):
    def __init__(self, compile_cache: Optional[CompileCache] = None):
        if compile_cache is None:
            compile_cache = CompileCache()

        self.rabi_ops = []
        self.detuning_ops = []
        self.level_coupling = None
        self.level_couplings = set()
        self.compile_cache = compile_cache

    def visit_emulator_program(self, emulator_program: EmulatorProgram):
        self.level_couplings = set(list(emulator_program.pulses.keys()))

        self.visit(emulator_program.register)
        for level_coupling, laser_coupling in emulator_program.pulses.items():
            self.level_coupling = level_coupling
            self.visit(laser_coupling)

    def visit_register(self, register: Register):
        self.register = register

        if register in self.compile_cache.space_cache:
            self.space, self.rydberg = self.compile_cache.space_cache[register]
            return

        self.space = Space.create(register)
        sites = register.sites

        # generate rydberg interaction elements
        self.rydberg = np.zeros(self.space.size, dtype=np.float64)

        for index_1, site_1 in enumerate(sites):
            site_1 = np.asarray(list(map(float, site_1)))
            is_rydberg_1 = self.space.is_rydberg_at(index_1)
            for index_2, sites_2 in enumerate(sites[index_1 + 1 :], index_1 + 1):
                sites_2 = np.asarray(list(map(float, sites_2)))
                distance = np.linalg.norm(site_1 - sites_2)

                rydberg_interaction = RB_C6 / (distance**6)

                if rydberg_interaction <= np.finfo(np.float64).eps:
                    continue

                mask = np.logical_and(is_rydberg_1, self.space.is_rydberg_at(index_2))
                self.rydberg[mask] += rydberg_interaction

        self.compile_cache.space_cache[register] = (self.space, self.rydberg)

    def visit_fields(self, fields: Fields):
        terms = fields.detuning + fields.rabi
        for term in terms:
            self.visit(term)

    def visit_detuning_operator_data(self, detuning_data: DetuningOperatorData):
        if (self.register, detuning_data) in self.compile_cache.operator_cache:
            return self.compile_cache.operator_cache[(self.register, detuning_data)]

        diagonal = np.zeros(self.space.size, dtype=np.float64)
        if self.space.atom_type == TwoLevelAtomType():
            state = TwoLevelAtomType.State.Rydberg
        elif self.space.atom_type == ThreeLevelAtomType():
            if self.level_coupling is LevelCoupling.Rydberg:
                state = ThreeLevelAtomType.State.Rydberg
            elif self.level_coupling is LevelCoupling.Hyperfine:
                state = ThreeLevelAtomType.State.Hyperfine

        for atom_index, value in detuning_data.target_atoms.items():
            diagonal[self.space.is_state_at(atom_index, state)] -= float(value)

        self.compile_cache.operator_cache[(self.register, detuning_data)] = diagonal
        return diagonal

    def visit_rabi_operator_data(self, rabi_operator_data: RabiOperatorData):
        if (self.register, rabi_operator_data) in self.compile_cache.operator_cache:
            return self.compile_cache.operator_cache[
                (self.register, rabi_operator_data)
            ]

        # get the form `to` and `from` states for the rabi term
        if self.space.atom_type == TwoLevelAtomType():
            to = TwoLevelAtomType.State.Ground
            fro = TwoLevelAtomType.State.Rydberg

        elif self.space.atom_type == ThreeLevelAtomType():
            if self.level_coupling is LevelCoupling.Rydberg:
                to = ThreeLevelAtomType.State.Hyperfine
                fro = ThreeLevelAtomType.State.Rydberg
            elif self.level_coupling == LevelCoupling.Hyperfine:
                to = ThreeLevelAtomType.State.Ground
                fro = ThreeLevelAtomType.State.Hyperfine

        # get matrix element generating function
        if rabi_operator_data.operator_type is RabiOperatorType.RabiSymmetric:

            def matrix_ele(atom_index):
                return self.space.swap_state_at(atom_index, fro, to)

        elif rabi_operator_data.operator_type is RabiOperatorType.RabiAsymmetric:

            def matrix_ele(atom_index):
                return self.space.transition_state_at(atom_index, fro, to)

        # generate rabi operator
        if len(rabi_operator_data.target_atoms) == 1:
            ((atom_index, _),) = rabi_operator_data.target_atoms.items()
            operator = IndexMapping(self.space.size, *matrix_ele(atom_index))
        else:
            indptr = np.zeros(self.space.size + 1, dtype=self.space.index_type)

            for atom_index in rabi_operator_data.target_atoms:
                row_indices, col_indices = matrix_ele(atom_index)
                indptr[1:][row_indices] += 1
            np.cumsum(indptr, out=indptr)

            indices = np.zeros(indptr[-1], dtype=self.space.index_type)
            data = np.zeros(indptr[-1], dtype=np.float64)

            for atom_index, value in rabi_operator_data.target_atoms.items():
                row_indices, col_indices = matrix_ele(atom_index)
                indices[indptr[:-1][row_indices]] = col_indices
                data[indptr[:-1][row_indices]] = value
                indptr[:-1][row_indices] += 1

            indptr[1:] = indptr[:-1]
            indptr[0] = 0

            operator = SparseMatrixCSR.create(
                (data, indices, indptr),
                shape=(self.space.size, self.space.size),
            )
            # operator = csr_matrix(
            #     (data, indices, indptr),
            #     shape=(self.space.size, self.space.size),
            # )

        self.compile_cache.operator_cache[
            (self.register, rabi_operator_data)
        ] = operator
        return operator

    def visit_detuning_term(self, detuning_term: DetuningTerm):
        self.detuning_ops.append(
            DetuningOperator(
                diagonal=self.visit(detuning_term.operator_data),
                amplitude=detuning_term.amplitude,
            )
        )

    def visit_rabi_term(self, rabi_term: RabiTerm):
        self.rabi_ops.append(
            RabiOperator(
                op=self.visit(rabi_term.operator_data),
                amplitude=rabi_term.amplitude,
                phase=rabi_term.phase,
            )
        )

    def emit(self, emulator_program: EmulatorProgram) -> RydbergHamiltonian:
        self.visit(emulator_program)
        hamiltonian = RydbergHamiltonian(
            emulator_ir=emulator_program,
            space=self.space,
            rydberg=self.rydberg,
            detuning_ops=self.detuning_ops,
            rabi_ops=self.rabi_ops,
        )
        return hamiltonian



from bloqade.ir.control.waveform import (
    AlignedWaveform,
    Constant,
    Linear,
    Poly,
    PythonFn,
    Sample,
)
from bloqade.ir.visitor.analog_circuit import AnalogCircuitVisitor
from bloqade.ir.visitor.waveform import WaveformVisitor
from bloqade.ir.analog_circuit import AnalogCircuit
import bloqade.ir.control.sequence as sequence
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.field as field
import bloqade.ir.control.waveform as waveform
import bloqade.ir.analog_circuit as analog_circuit

import numbers
from typing import Any, Dict


class AssignmentScanRecord(WaveformVisitor):
    def __init__(self, assignments: Dict[str, numbers.Real] = {}):
        self.assignments = dict(assignments)

    def visit_record(self, ast: waveform.Record):
        self.visit(ast.waveform)
        duration = ast.waveform.duration(**self.assignments)
        var = ast.var
        value = ast.waveform.eval_decimal(duration, **self.assignments)
        self.assignments[var.name] = value

    def visit_append(self, ast: waveform.Append):
        list(map(self.visit, ast.waveforms))

    def visit_slice(self, ast: waveform.Slice):
        self.visit(ast.waveform)

    def visit_add(self, ast: waveform.Add):
        self.visit(ast.left)
        self.visit(ast.right)

    def visit_negative(self, ast: waveform.Negative):
        self.visit(ast.waveform)

    def visit_scale(self, ast: waveform.Scale):
        self.visit(ast.waveform)

    def visit_smooth(self, ast: waveform.Smooth):
        self.visit(ast.waveform)

    def visit_sample(self, ast: Sample) -> Any:
        return self.visit(ast.waveform)

    def visit_alligned(self, ast: AlignedWaveform) -> Any:
        return super().visit_alligned(ast)

    def visit_constant(self, ast: Constant) -> Any:
        pass

    def visit_linear(self, ast: Linear) -> Any:
        pass

    def visit_poly(self, ast: Poly) -> Any:
        pass

    def visit_python_fn(self, ast: PythonFn) -> Any:
        pass

    def emit(self, ast: waveform.Waveform) -> Dict[str, numbers.Real]:
        self.visit(ast)
        return self.assignments


class AssignmentScan(AnalogCircuitVisitor):
    def __init__(self, assignments: Dict[str, numbers.Real] = {}):
        self.assignments = dict(assignments)
        self.waveform_visitor = AssignmentScanRecord(self.assignments)

    def visit_analog_circuit(self, ast: AnalogCircuit) -> Any:
        self.visit(ast.sequence)

    def visit_sequence(self, ast: sequence.Sequence):
        list(map(self.visit, ast.pulses.values()))

    def visit_named_sequence(self, ast: sequence.NamedSequence):
        self.visit(ast.sequence)

    def visit_append_sequence(self, ast: sequence.Append):
        list(map(self.visit, ast.sequences))

    def visit_slice_sequence(self, ast: sequence.Slice):
        self.visit(ast.sequence)

    def visit_pulse(self, ast: pulse.Pulse):
        list(map(self.visit, ast.fields.values()))

    def visit_named_pulse(self, ast: pulse.NamedPulse) -> Any:
        self.visit(ast.pulse)

    def visit_append_pulse(self, ast: pulse.Append) -> Any:
        list(map(self.visit, ast.pulses))

    def visit_slice_pulse(self, ast: pulse.Slice) -> Any:
        self.visit(ast.pulse)

    def visit_field(self, ast: field.Field):
        list(map(self.visit, ast.drives.values()))

    def visit_waveform(self, ast: waveform.Waveform):
        self.assignments.update(self.waveform_visitor.emit(ast))

    def emit(self, ast: analog_circuit.AnalogCircuit) -> Dict[str, numbers.Real]:
        self.visit(ast)
        return self.assignments



from pydantic import BaseModel, validator, ValidationError

from typing import List, Optional, Tuple
from itertools import combinations


class ClusterLocationInfo(BaseModel):
    """Class that stores the mapping of batched jobs.

    Args:
        cluster_index (int): the index of the cluster a site belongs to
        global_location_index (int): the index of the site in the multplexed system
        cluster_location_index (int): the index of the site in the original system
    """

    cluster_index: Tuple[int, int]
    global_location_index: int
    cluster_location_index: int


class ParallelDecoder(BaseModel):
    mapping: List[ClusterLocationInfo]
    locations_per_cluster: int
    number_of_cluster: int

    class Config:
        frozen = True

    def __init__(
        self,
        mapping: List[ClusterLocationInfo],
        locations_per_cluster: Optional[int] = None,
        number_of_cluster: Optional[int] = None,
    ):
        if locations_per_cluster is None:
            cluster_location_indices = set()

            for site in mapping:
                cluster_location_indices.add(site.cluster_location_index)

            locations_per_cluster = len(cluster_location_indices)

        if number_of_cluster is None:
            cluster_indices = set()

            for site in mapping:
                cluster_indices.add(site.cluster_index)

            number_of_cluster = len(cluster_indices)

        super().__init__(
            mapping=mapping,
            locations_per_cluster=number_of_cluster,
            number_of_cluster=number_of_cluster,
        )

    # should work if we go to the coordinate-based indexing system
    @validator("mapping", allow_reuse=True)
    def sites_belong_to_unqiue_cluster(cls, mapping):
        sites = [ele.global_location_index for ele in mapping]
        unique_sites = set(sites)
        if len(sites) != len(unique_sites):
            raise ValidationError("one or more sites mapped to multiple clusters")

        cluster_indices = {}
        for location_cluster_info in mapping:
            cluster_index = location_cluster_info.cluster_index
            global_site_index = location_cluster_info.global_location_index
            cluster_indices.setdefault(cluster_index, set([])).add(global_site_index)

        for (cluster_index, location_indices), (
            other_cluster_index,
            other_location_indices,
        ) in combinations(cluster_indices.items(), 2):
            if location_indices.intersection(other_location_indices):
                raise ValidationError("one or more sites mapped to multiple clusters")

        return mapping

    # map individual atom indices (in the context of the ENTIRE geometry)
    # to the cluster-specific indices:
    # {}
    def get_location_indices(self):
        site_indices = {}
        for location_cluster_info in self.mapping:  # iterate through global_site_index
            global_site_index = location_cluster_info.global_location_index
            local_site_index = location_cluster_info.cluster_location_index
            site_indices[global_site_index] = local_site_index

        return site_indices

    # should work if we go to coordinate-based indexing
    # map each cluster index to the global index
    def get_cluster_indices(self):
        site_indices = self.get_location_indices()

        cluster_indices = {}
        for location_cluster_info in self.mapping:
            global_site_index = location_cluster_info.global_location_index
            cluster_index = location_cluster_info.cluster_index

            cluster_indices[cluster_index] = cluster_indices.get(cluster_index, []) + [
                global_site_index
            ]

        return {
            cluster_index: sorted(sites, key=lambda site: site_indices[site])
            for cluster_index, sites in cluster_indices.items()
        }



from bloqade.ir import (
    Field,
    Uniform,
    Linear,
    ScaledLocations,
    Location,
    SpatialModulation,
    RunTimeVector,
)
from bloqade.ir.control.field import Drive
import pytest
from bloqade import cast
from io import StringIO
from IPython.lib.pretty import PrettyPrinter as PP

import bloqade.ir.tree_print as trp

trp.color_enabled = False


def test_location():
    loc = Location(3)

    assert str(loc) == "Location(3)"
    assert loc.print_node() == "Location(3)"
    assert loc.children() == []


def test_spacmod_base():
    x = SpatialModulation()

    with pytest.raises(NotImplementedError):
        hash(x)


def test_unform():
    x = Uniform

    assert str(x) == "UniformModulation\n"
    assert x.print_node() == "UniformModulation"
    assert x.children() == []

    mystdout = StringIO()
    p = PP(mystdout)

    x._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "UniformModulation\n"


def test_runtime_vec():
    x = RunTimeVector("sss")

    assert str(x) == "RunTimeVector\n" + "└─ sss\n"
    assert x.print_node() == "RunTimeVector"
    assert x.children() == ["sss"]

    mystdout = StringIO()
    p = PP(mystdout)

    x._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "RunTimeVector\n" + "└─ sss\n"


def test_scal_loc():
    x = ScaledLocations({1: 1.0, 2: 2.0})

    with pytest.raises(ValueError):
        ScaledLocations({(2, 3): 2})

    assert x.print_node() == "ScaledLocations"
    assert x.children() == {"Location(1)": cast(1.0), "Location(2)": cast(2.0)}

    mystdout = StringIO()
    p = PP(mystdout)

    x._repr_pretty_(p, 0)

    assert mystdout.getvalue() == (
        "ScaledLocations\n"
        "├─ Location(1)\n"
        "│  ⇒ Literal: 1.0\n"
        "⋮\n"
        "└─ Location(2)\n"
        "   ⇒ Literal: 2.0⋮\n"
    )


def test_field_scaled_locations():
    Loc = ScaledLocations({1: 1.0, 2: 2.0})
    Loc2 = ScaledLocations({3: 1.0, 4: 2.0})
    Loc3 = ScaledLocations({1: 1.0, 2: 2.0, 3: 1.0, 4: 2.0})
    f1 = Field({Loc: Linear(start=1.0, stop="x", duration=3.0)})
    f2 = Field({Loc: Linear(start=1.0, stop="x", duration=3.0)})
    f3 = Field({Loc2: Linear(start=1.0, stop="x", duration=3.0)})
    f4 = Field({Loc3: Linear(start="y", stop="x", duration=3.0)})

    # add with non field
    with pytest.raises(ValueError):
        f1.add(Loc)

    mystdout = StringIO()
    p = PP(mystdout)

    f1._repr_pretty_(p, 10)

    assert mystdout.getvalue() == (
        "Field\n"
        "└─ Drive\n"
        "   ├─ modulation\n"
        "   │  ⇒ ScaledLocations\n"
        "   │    ├─ Location(1)\n"
        "   │    │  ⇒ Literal: 1.0\n"
        "   │    └─ Location(2)\n"
        "   │       ⇒ Literal: 2.0\n"
        "   └─ waveform\n"
        "      ⇒ Linear\n"
        "        ├─ start\n"
        "        │  ⇒ Literal: 1.0\n"
        "        ├─ stop\n"
        "        │  ⇒ Variable: x\n"
        "        └─ duration\n"
        "           ⇒ Literal: 3.0"
    )

    # add with field same spat-mod
    o1 = f1.add(f2)
    assert len(o1.drives.keys()) == 1

    # add with field diff spat-mod
    o2 = f1.add(f3)
    assert o2 == Field({Loc3: Linear(start=1.0, stop="x", duration=3.0)})
    # assert len(o2.drives.keys()) == 1

    assert f2.print_node() == "Field"
    # assert type(hash(f1)) == int
    assert f1.children() == [Drive(k, v) for k, v in f1.drives.items()]

    assert hash(f1)

    o3 = f1.add(f4)
    assert o3 == Field(
        {
            Loc3: Linear(start="y", stop="x", duration=3.0),
            Loc: Linear(start=1.0, stop="x", duration=3.0),
        }
    )



from bloqade.serialize import Serializer
from bloqade.builder.base import ParamType
from .base import LocalTask
from bloqade.submission.ir.task_results import QuEraTaskResults
from bloqade.submission.ir.braket import (
    from_braket_task_results,
    BraketTaskSpecification,
)
from bloqade.task.base import Geometry
from braket.devices import LocalSimulator
from beartype.typing import Dict, Optional, Any
from dataclasses import dataclass

## keep the old conversion for now,
## we will remove conversion btwn QuEraTask <-> BraketTask,
## and specialize/dispatching here.


@dataclass
@Serializer.register
class BraketEmulatorTask(LocalTask):
    task_ir: BraketTaskSpecification
    metadata: Dict[str, ParamType]
    task_result_ir: Optional[QuEraTaskResults] = None

    def _geometry(self) -> Geometry:
        return Geometry(
            sites=self.task_ir.program.setup.ahs_register.sites,
            filling=self.task_ir.program.setup.ahs_register.filling,
        )

    def run(self, **kwargs) -> "BraketEmulatorTask":
        aws_task = LocalSimulator("braket_ahs").run(
            self.task_ir.program,
            shots=self.task_ir.nshots,
            **kwargs,
        )
        self.task_result_ir = from_braket_task_results(aws_task.result())
        return self

    def result(self):
        if self.task_result_ir is None:
            raise ValueError("Braket simulator job haven't submit yet.")

        return self.task_result_ir


@BraketEmulatorTask.set_serializer
def _serialize(obj: BraketEmulatorTask) -> Dict[str, Any]:
    return {
        "task_ir": obj.task_ir.dict(),
        "metadata": obj.metadata,
        "task_result_ir": obj.task_result_ir.dict() if obj.task_result_ir else None,
    }


@BraketEmulatorTask.set_deserializer
def _serializer(d: Dict[str, Any]) -> BraketEmulatorTask:
    d["task_ir"] = BraketTaskSpecification(**d["task_ir"])
    d["task_result_ir"] = (
        QuEraTaskResults(**d["task_result_ir"]) if d["task_result_ir"] else None
    )
    return BraketEmulatorTask(**d)



from juliacall import Main, TypeValue  # type: ignore
from typing import List, Dict, Any
from dataclasses import dataclass
from .to_julia import ToJulia

PythonCall = Main.seval("PythonCall")


@dataclass(frozen=True)
class JLTypeVar:
    name: str


class JLType:
    """Superclass for all Julia types."""

    def __init__(self, expr, *typeparams) -> None:
        self.expr = expr
        self.obj = getattr(Main, expr)
        params = []
        for param in typeparams:
            if isinstance(param, JLType):
                params.append(param.obj)
            elif isinstance(param, TypeValue):
                params.append(param)
            else:
                raise Exception(f"Unknown type parameter: {param}")
        self.typeparams = params

    def __call__(self, x) -> Any:
        if self.typeparams:
            jl_type = self.obj[*self.typeparams]
        else:
            jl_type = self.obj
        return PythonCall.pyconvert(jl_type, x)

    def __getitem__(self, *typeparams: Any) -> Any:
        return JLType(self.expr, *typeparams)

    def __repr__(self) -> str:
        repr = self.expr
        if self.typeparams:
            repr += "[" + ",".join(map(str, self.typeparams)) + "]"
        return repr


class JLVectorType(JLType):
    def __init__(self, expr, *typeparams) -> None:
        super().__init__(expr, *typeparams)

    def __call__(self, x: List[ToJulia]) -> Any:
        type = self.obj[*self.typeparams]
        ret = type(Main.undef, len(x))
        for i, v in enumerate(x):
            ret[i] = v.julia()
        return ret

    def __getitem__(self, *typeparams: Any) -> Any:
        assert len(typeparams) == 1
        assert self.typeparams == []
        return JLVectorType(*typeparams)


class JLDictType(JLType):
    def __init__(self, expr, *typeparams) -> None:
        super().__init__(expr, *typeparams)

    def __call__(self, x: Dict[ToJulia, ToJulia]) -> Any:
        type = self.obj[*self.typeparams]
        ret = type()
        for k, v in x.items():
            ret[k.julia()] = v.julia()
        return ret

    def __getitem__(self, *typeparams: Any) -> Any:
        assert len(self.typeparams) + len(typeparams) == 2
        return JLDictType(self.expr, *typeparams)


Int32 = JLType("Int32")
Int64 = JLType("Int64")
Float32 = JLType("Float32")
Float64 = JLType("Float64")
Complex = JLType("Complex")
ComplexF32 = JLType("ComplexF32")
ComplexF64 = JLType("ComplexF64")
String = JLType("String")
Bool = JLType("Bool")
Symbol = JLType("Symbol")
Vector = JLVectorType("Vector")
Dict = JLDictType("Dict")



from beartype.typing import Optional
from beartype import beartype
from bloqade.builder.typing import LiteralType
from bloqade.builder.base import Builder
from bloqade.builder.backend import BackendRoute
from bloqade.ir import cast


class Parallelize(BackendRoute, Builder):
    @beartype
    def __init__(
        self, cluster_spacing: LiteralType, parent: Optional[Builder] = None
    ) -> None:
        super().__init__(parent)
        self._cluster_spacing = cast(cluster_spacing)



from beartype.typing import Optional, TYPE_CHECKING

from beartype import beartype
from beartype.typing import List, Union
from bloqade.builder.typing import ScalarType, LiteralType
from bloqade.builder.waveform import WaveformAttachable
from bloqade.builder.base import Builder


if TYPE_CHECKING:
    from bloqade.ir.control.field import (
        UniformModulation,
        ScaledLocations,
        AssignedRunTimeVector,
        RunTimeVector,
    )


class SpatialModulation(WaveformAttachable):
    pass


class Uniform(SpatialModulation):
    """
    The node specify a uniform spacial modulation. Which is ready to apply waveform
    (See [`Waveform`][bloqade.builder.waveform] for available waveform options)

    Examples:

        - To hit this node from the start node:

        >>> reg = bloqade.start.add_position([(0,0),(1,1),(2,2),(3,3)])
        >>> loc = reg.rydberg.detuning.uniform

        - Apply Linear waveform:

        >>> wv = bloqade.ir.Linear(start=0,stop=1,duration=0.5)
        >>> reg = bloqade.start.add_position([(0,0),(1,1),(2,2),(3,3)])
        >>> loc = reg.rydberg.detuning.uniform.apply(wv)

    """

    def __bloqade_ir__(self) -> "UniformModulation":
        from bloqade.ir import Uniform

        return Uniform


class Location(SpatialModulation):
    @beartype
    def __init__(
        self,
        labels: List[int],
        scales: List[ScalarType],
        parent: Optional[Builder] = None,
    ) -> None:
        from bloqade.ir.scalar import cast

        super().__init__(parent)
        self._scaled_locations = {
            label: cast(scale) for label, scale in zip(labels, scales)
        }

    def __bloqade_ir__(self) -> "ScaledLocations":
        from bloqade.ir import ScaledLocations

        return ScaledLocations(self._scaled_locations)


class Scale(SpatialModulation):
    @beartype
    def __init__(
        self,
        name_or_list: Union[str, List[LiteralType]],
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._name_or_list = name_or_list

    def __bloqade_ir__(self) -> Union["RunTimeVector", "AssignedRunTimeVector"]:
        from bloqade.ir import RunTimeVector, AssignedRunTimeVector

        if isinstance(self._name_or_list, str):
            return RunTimeVector(self._name_or_list)
        else:
            return AssignedRunTimeVector(None, self._name_or_list)



# from bloqade import start
from bloqade.ir import location, cast, Interpolation
from bloqade.ir import Constant, Linear, Poly, PythonFn, Sample
import pytest
import json
import numpy as np


"""
import bloqade.ir.location as location
from bloqade.codegen.hardware.quera import SchemaCodeGen
from bloqade.submission.base import get_capabilities

bloqade_program = (
    location.Square(1)
    .rydberg.rabi.phase.uniform.piecewise_constant(
        durations=[0.5, 0.5], values=[0, 1]
    )
    .piecewise_constant(
        durations=[0.3] ,values = [0.2]
    ).program
)

capabilities = get_capabilities()
schema = SchemaCodeGen({},capabilities=capabilities).emit(10, bloqade_program)

print(schema.effective_hamiltonian.rydberg.rabi_frequency_phase.global_.times)
"""


def fc(x):
    return float(x)


def fvec(v):
    return list(map(fc, v))


def test_integration_jump_err():
    ## jump at the end of linear -- constant
    with pytest.raises(ValueError):
        (
            location.Square(6)
            .rydberg.detuning.uniform.apply(
                Constant("initial_detuning", "up_time")
                .append(Linear("initial_detuning", "final_detuning", "anneal_time"))
                .append(0.5 * Constant("final_detuning", "up_time"))
            )
            .assign(
                initial_detuning=-10,
                up_time=0.1,
                final_detuning=15,
                anneal_time=10,
            )
            .quera.mock()
            ._compile(10)
        )


def test_integration_scale():
    seq = Linear(start=0.0, stop=1.0, duration=0.5).append(
        2 * Constant(0.5, duration=0.5)
    )
    job = (
        location.Square(1)
        .rydberg.detuning.uniform.apply(seq)
        .quera.mock()
        ._compile(shots=10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([0, 1, 1]) * 1e6)


def test_integration_neg():
    seq = Linear(start=0.0, stop=-0.5, duration=0.5).append(
        -Constant(0.5, duration=0.5)
    )
    job = (
        location.Square(1)
        .rydberg.detuning.uniform.apply(seq)
        .quera.mock()
        ._compile(shots=10)
    )

    panel = json.loads(job.json())
    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([0, -0.5, -0.5]) * 1e6)


def test_integration_poly_order_err():
    ## poly
    with pytest.raises(ValueError):
        seq = Poly(coeffs=[1, 2, 3], duration=0.5).append(-Constant(0.5, duration=0.5))
        (
            location.Square(1)
            .rydberg.detuning.uniform.apply(seq)
            .quera.mock()
            ._compile(shots=10)
        )


def test_integration_poly_const():
    ## constant
    seq = Poly(coeffs=[1], duration=0.5).append(Constant(1, duration=0.5))
    job = (
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(12)
    )

    panel = json.loads(job.json())
    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 12
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([1, 1, 1]) * 1e6)


def test_integration_poly_linear():
    ## linear
    seq = Poly(coeffs=[1, 2], duration=0.5).append(Constant(2, duration=0.5))
    job = (
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)
    )

    panel = json.loads(job.json())
    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([1, 2, 2]) * 1e6)


def test_integration_linear_sampl_const_err():
    def my_cos(time):
        return np.cos(time)

    assert my_cos(1) == np.cos(1)

    wv = PythonFn.create(my_cos, duration=1.0)
    dt = cast(0.1)

    wf = Sample(wv, Interpolation.Constant, dt)
    ## phase can only have piecewise constant.
    with pytest.raises(ValueError):
        location.Square(1).rydberg.detuning.uniform.apply(wf).quera.mock()._compile(10)


def test_integration_slice_linear_const():
    seq = Linear(start=0.0, stop=1.0, duration=1.0)[0:0.5].append(
        Constant(0.5, duration=0.5)
    )
    job = (
        location.Square(1)
        .rydberg.detuning.uniform.apply(seq)
        .quera.mock()
        ._compile(shots=10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([0, 0.5, 0.5]) * 1e6)


def test_integration_slice_linear_no_stop():
    seq = Linear(start=0.0, stop=1.0, duration=1.0)[0:]
    job = (
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([0, 1.0]) * 1e6)


def test_integration_slice_linear_no_start():
    seq = Linear(start=0.0, stop=1.0, duration=1.0)[:1.0]
    job = (
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]
    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([0, 1.0]) * 1e6)


def test_integration_slice_linear_error_neg_start():
    with pytest.raises(ValueError):
        seq = Linear(start=0.0, stop=1.0, duration=1.0)[-0.1:1.0]
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)


def test_integration_slice_linear_error_exceed_stop():
    with pytest.raises(ValueError):
        seq = Linear(start=0.0, stop=1.0, duration=1.0)[:4.0]
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)


def test_integration_slice_linear_error_revese():
    with pytest.raises(ValueError):
        seq = Linear(start=0.0, stop=1.0, duration=1.0)[0.5:0.0]
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)


def test_integration_slice_linear_error_same_vals_nofield():
    with pytest.raises(ValueError):
        seq = Linear(start=0.0, stop=1.0, duration=1.0)[0.0:0.0]
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)


def test_integration_phase():
    batch = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.piecewise_constant(
            durations=[0.5, 0.5], values=[0, 1]
        )
        .record("a")
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(batch.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([0, 1.0, 1.0]))


def test_intergration_phase_nonconst_err():
    with pytest.raises(ValueError):
        (
            location.Square(1)
            .rydberg.rabi.phase.uniform.piecewise_linear(
                durations=[0.5, 0.5], values=[0, 1, 2]
            )
            .quera.mock()
            ._compile(10)
        )

    seq = Poly(coeffs=[1, 2], duration=0.5)
    with pytest.raises(ValueError):
        location.Square(1).rydberg.rabi.phase.uniform.apply(seq).quera.mock()._compile(
            10
        )


def test_integration_phase_linear():
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.linear(start=1, stop=1, duration=1)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 1.0]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([1.0, 1.0]))


def test_integration_phase_polyconst():
    seq = Poly(coeffs=[1], duration=0.5)
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([1.0, 1.0]))


def test_integration_phase_slice():
    ##[Further investigate!]
    seq = Poly(coeffs=[1], duration=1.0)[0:0.5]
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([1.0, 1.0]))


def test_integration_phase_scale():
    seq = 3.0 * Constant(value=1.0, duration=1.0)
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 1.0]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([3.0, 3.0]))


def test_integration_phase_neg():
    seq = -Constant(value=1.0, duration=1.0)
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 1.0]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([-1.0, -1.0]))


def test_integration_phase_slice_no_start():
    seq = Poly(coeffs=[1], duration=1.0)[:0.5]
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([1.0, 1.0]))


def test_integration_phase_slice_no_stop():
    ##[Further investigate!]
    seq = Poly(coeffs=[1], duration=0.5)[0:]
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([1.0, 1.0]))


def test_integration_phase_slice_same_start_stop():
    ##[Further investigate!]
    seq = Poly(coeffs=[1], duration=0.5)[0:0.5]
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([1.0, 1.0]))


def test_integration_phase_slice_error_neg_start():
    with pytest.raises(ValueError):
        seq = Poly(coeffs=[1], duration=1.0)[-0.1:0.5]
        location.Square(1).rydberg.rabi.phase.uniform.apply(seq).quera.mock()._compile(
            10
        )


def test_integration_phase_slice_error_exceed_stop():
    with pytest.raises(ValueError):
        seq = Poly(coeffs=[1], duration=1.0)[0:2.0]
        location.Square(1).rydberg.rabi.phase.uniform.apply(seq).quera.mock()._compile(
            10
        )


def test_integration_phase_slice_error_reverse():
    with pytest.raises(ValueError):
        seq = Poly(coeffs=[1], duration=1.0)[2.0:0.0]
        location.Square(1).rydberg.rabi.phase.uniform.apply(seq).quera.mock()._compile(
            10
        )


def test_integration_phase_sampl_linear_err():
    def my_cos(time):
        return np.cos(time)

    assert my_cos(1) == np.cos(1)

    wv = PythonFn.create(my_cos, duration=1.0)
    dt = cast(0.1)

    wf = Sample(wv, Interpolation.Linear, dt)
    ## phase can only have piecewise constant.
    with pytest.raises(ValueError):
        location.Square(1).rydberg.rabi.phase.uniform.apply(wf).quera.mock()._compile(
            10
        )


def test_integration_batchassign_assign():
    ## jump at the end of linear -- constant
    with pytest.raises(ValueError):
        (
            location.Square(6)
            .rydberg.detuning.uniform.apply(Constant("initial_detuning", "time"))
            .batch_assign(initial_detuning=[1, 2, 3, 4], time=[2, 4, 5])
            .quera.mock()
            ._compile(10)
        )

    ## jump at the end of linear -- constant

    batch = (
        location.Square(6)
        .rydberg.detuning.uniform.apply(Constant("initial_detuning", "time"))
        .batch_assign(initial_detuning=[1, 2, 3, 4], time=[2, 4, 5, 0.6])
        .quera.mock()
        ._compile(10)
    )

    assert len(batch.tasks) == 4


def test_integration_record():
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.piecewise_constant(
            durations=[0.5, 0.5], values=[0, 1]
        )
        .record("a")
        .piecewise_constant(durations=[0.3], values=["a"])
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5, 1.0, 1.3]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([0, 1.0, 1.0, 1.0]))


pytest.mark.skip(reason="fixing now")


def test_integration_fn_phase():
    def my_cos(time):
        return np.cos(time)

    assert my_cos(1) == np.cos(1)

    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.fn(my_cos, duration=1.0)
        .sample(dt=0.5)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(
        fvec(phase_ir["global"]["values"]) == np.array([1.0, 0.8775825, 0.8775825])
    )


pytest.mark.skip(reason="fixing now")


def test_integration_fn_detune():
    def my_cos(time):
        return np.cos(time)

    assert my_cos(1) == np.cos(1)

    job = (
        location.Square(1)
        .rydberg.detuning.uniform.fn(my_cos, duration=1.0)
        .sample(dt=0.5)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(
        fvec(detune_ir["global"]["values"]) == np.array([1000000, 877582.6, 540302.4])
    )



from collections import OrderedDict

import json
from typing import List, TextIO, Type, TypeVar, Union, Dict, Optional, Tuple
from numbers import Number

from pydantic import BaseModel
from bloqade.submission.ir.task_results import (
    QuEraTaskResults,
    QuEraTaskStatusCode,
)
from numpy.typing import NDArray
import pandas as pd
import numpy as np
from pydantic.dataclasses import dataclass
from bloqade.submission.ir.parallel import ParallelDecoder
import datetime
from bloqade.visualization import display_report

JSONSubType = TypeVar("JSONSubType", bound="JSONInterface")


@dataclass(frozen=True)
class Geometry:
    sites: List[Tuple[float, float]]
    filling: List[int]
    parallel_decoder: Optional[ParallelDecoder] = None


class JSONInterface(BaseModel):
    def json(self, exclude_none=True, by_alias=True, **json_options) -> str:
        return super().json(
            exclude_none=exclude_none, by_alias=by_alias, **json_options
        )

    def save_json(
        self, filename_or_io: Union[str, TextIO], mode="w", **json_options
    ) -> None:
        if isinstance(filename_or_io, str):
            with open(filename_or_io, mode) as f:
                f.write(self.json(**json_options))
        else:
            filename_or_io.write(self.json(**json_options))

    @classmethod
    def load_json(
        cls: Type[JSONSubType], filename_or_io: Union[str, TextIO]
    ) -> JSONSubType:
        if isinstance(filename_or_io, str):
            with open(filename_or_io, "r") as f:
                params = json.load(f)
        else:
            params = json.load(filename_or_io)

        return cls(**params)


class Task:
    def _geometry(self) -> Geometry:
        raise NotImplementedError(
            f"{self.__class__.__name__}._geometry() not implemented"
        )

    @property
    def geometry(self) -> Geometry:
        return self._geometry()


class RemoteTask(Task):
    def validate(self) -> None:
        raise NotImplementedError

    def result(self) -> QuEraTaskResults:
        # online, Blocking
        # waiting for remote results to finish
        # return results
        raise NotImplementedError

    def fetch(self) -> None:
        # online, non-blocking
        # pull the result if they are complete
        raise NotImplementedError

    def status(self) -> QuEraTaskStatusCode:
        # online, non-blocking
        # probe current task status
        raise NotImplementedError

    def pull(self) -> None:
        # online, blocking
        # pull the current results
        raise NotImplementedError

    def cancel(self) -> None:
        # this method cancels the task
        raise NotImplementedError

    def submit(self, force: bool):
        # online, non-blocking
        # this method submit the task
        raise NotImplementedError

    def _result_exists(self) -> bool:
        raise NotImplementedError


class LocalTask(Task):
    def result(self):
        # need a new results type
        # for emulator jobs
        raise NotImplementedError

    def run(self, **kwargs):
        raise NotImplementedError


# Report is now just a helper class for
# organize and analysis data:
class Report:
    dataframe: pd.DataFrame
    metas: List[Dict]
    geos: List[Geometry]
    name: str = ""

    def __init__(self, data, metas, geos, name="") -> None:
        self.dataframe = data  # df
        self._bitstrings = None  # bitstring cache
        self._counts = None  # counts cache
        self.metas = metas
        self.geos = geos
        self.name = name + " " + str(datetime.datetime.now())

    def list_param(self, field_name: str) -> List[Union[Number, None]]:
        """
        List the parameters associate with the given variable field_name
        for each tasks.

        Args:
            field_name (str): variable name

        """

        def cast(x):
            try:
                return float(x)
            except ValueError:
                return x

        return list(map(cast, (meta.get(field_name) for meta in self.metas)))

    @property
    def markdown(self) -> str:
        return self.dataframe.to_markdown()

    def bitstrings(self, filter_perfect_filling: bool = True) -> List[NDArray]:
        """
        Get the bitstrings from the data.

        Args:
            filter_perfect_filling (bool): whether return will
            only contain perfect filling shots.

        """
        perfect_sorting = self.dataframe.index.get_level_values("perfect_sorting")
        pre_sequence = self.dataframe.index.get_level_values("pre_sequence")
        if filter_perfect_filling:
            df = self.dataframe[perfect_sorting == pre_sequence]
        else:
            df = self.dataframe

        task_numbers = df.index.get_level_values("task_number").unique()

        bitstrings = []
        for task_number in task_numbers:
            bitstrings.append(df.loc[task_number, ...].to_numpy())

        return bitstrings

    @property
    def counts(self) -> List[OrderedDict[str, int]]:
        """
        Get the statistic for the counts for each configuration(s)

        Return:
            statistic of configuration (str): counts (int) for each task

        """
        if self._counts is not None:
            return self._counts
        self._counts = self._construct_counts()
        return self._counts

    def _construct_counts(self) -> List[OrderedDict[str, int]]:
        counts = []
        for bitstring in self.bitstrings():
            output = np.unique(bitstring, axis=0, return_counts=True)

            count_list = [
                ("".join(map(str, bitstring)), count)
                for bitstring, count in zip(*output)
            ]
            count_list.sort(key=lambda x: x[1], reverse=True)
            count = OrderedDict(count_list)

            counts.append(count)

        return counts

    def rydberg_densities(self, filter_perfect_filling: bool = True) -> pd.Series:
        """
        Get rydberg density for each task.

        Args:
            filter_perfect_filling (bool):  whether return will
            only contain perfect filling shots.

        Return:
            per-site rydberg density for each task

        """
        # TODO: implement nan for missing task numbers
        perfect_sorting = self.dataframe.index.get_level_values("perfect_sorting")
        pre_sequence = self.dataframe.index.get_level_values("pre_sequence")

        if filter_perfect_filling:
            df = self.dataframe[perfect_sorting == pre_sequence]
        else:
            df = self.dataframe

        return 1 - (df.groupby("task_number").mean())

    def show(self):
        """
        Interactive Visualization of the Report

        """
        display_report(self)



from enum import Enum
from pydantic import BaseModel, conlist, conint
from typing import Callable, Optional, List, Tuple
import numpy as np

__all__ = ["QuEraTaskResults", "TaskProbabilities"]

# TODO: add version to these models.


class QuEraShotStatusCode(str, Enum):
    Completed = "Completed"
    MissingPreSequence = "MissingPreSequence"
    MissingPostSequence = "MissingPostSequence"
    MissingMeasurement = "MissingMeasurement"


class QuEraTaskStatusCode(str, Enum):
    Created = "Created"
    Running = "Running"
    Completed = "Completed"
    Failed = "Failed"
    Cancelled = "Cancelled"


class QuEraShotResult(BaseModel):
    shot_status: QuEraShotStatusCode = QuEraShotStatusCode.MissingMeasurement
    pre_sequence: conlist(conint(ge=0, le=1), min_items=0) = []
    post_sequence: conlist(conint(ge=0, le=1), min_items=0) = []


class TaskProbabilities(BaseModel):
    probabilities: List[Tuple[Tuple[str, str], float]]

    def simulate_task_results(self, shots=1) -> "QuEraTaskResults":
        bit_strings, probabilities = zip(*self.probabilities)

        indices = np.random.choice(len(probabilities), p=probabilities, size=shots)
        shot_outputs = []
        for index in indices:
            pre_string, post_string = bit_strings[index]
            pre_sequence = [int(bit) for bit in pre_string]
            post_sequence = [int(bit) for bit in post_string]

            shot_outputs.append(
                QuEraShotResult(
                    shot_status=QuEraShotStatusCode.Completed,
                    pre_sequence=pre_sequence,
                    post_sequence=post_sequence,
                )
            )

        return QuEraTaskResults(
            task_status=QuEraTaskStatusCode.Completed, shot_outputs=shot_outputs
        )


class QuEraTaskResults(BaseModel):
    task_status: QuEraTaskStatusCode = QuEraTaskStatusCode.Failed
    shot_outputs: conlist(QuEraShotResult, min_items=0) = []

    def export_as_probabilities(self) -> TaskProbabilities:
        """converts from shot results to probabilities

        Returns:
            TaskProbabilities: The task results as probabilties
        """
        counts = dict()
        nshots = len(self.shot_outputs)
        for shot_result in self.shot_outputs:
            pre_sequence_str = "".join(str(bit) for bit in shot_result.pre_sequence)

            post_sequence_str = "".join(str(bit) for bit in shot_result.post_sequence)

            configuration = (pre_sequence_str, post_sequence_str)
            # iterative average
            current_count = counts.get(configuration, 0)
            counts[configuration] = current_count + 1

        probabilities = [(config, count / nshots) for config, count in counts.items()]
        return TaskProbabilities(probabilities=probabilities)

    def post_process(
        self, keep_shot_result: Optional[Callable] = None, args=()
    ) -> "QuEraTaskResults":
        if keep_shot_result is None:

            def filter_func(shot_result):
                return all(bit == 1 for bit in shot_result.pre_sequence)

        else:

            def filter_func(shot_result):
                keep_shot_result(shot_result, *args)

        return QuEraTaskResults(
            task_status=self.task_status,
            shot_outputs=list(filter(filter_func, self.shot_outputs)),
        )



# from bloqade.ir.routine.params import Params
from bloqade.ir.analog_circuit import AnalogCircuit
from bloqade.ir import AtomArrangement, ParallelRegister, Sequence

from bloqade.builder.base import Builder
from bloqade.builder.parse.trait import Parse, Show
from bloqade.ir.routine.params import Params

from pydantic import ConfigDict
from pydantic.dataclasses import dataclass
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from bloqade.ir.routine.braket import BraketServiceOptions
    from bloqade.ir.routine.quera import QuEraServiceOptions
    from bloqade.ir.routine.bloqade import BloqadeServiceOptions


class RoutineParse(Parse):
    def parse_register(self: "RoutineBase") -> Union[AtomArrangement, ParallelRegister]:
        return self.circuit.register

    def parse_sequence(self: "RoutineBase") -> Sequence:
        return self.circuit.sequence

    def parse_circuit(self: "RoutineBase") -> AnalogCircuit:
        return self.circuit

    def parse(self: "RoutineBase") -> "Routine":
        return self


class RoutineShow(Show):
    def show(self: "RoutineBase", batch_index: int = 0):
        return self.source.show(batch_index)


__pydantic_dataclass_config__ = ConfigDict(arbitrary_types_allowed=True)


@dataclass(frozen=True, config=__pydantic_dataclass_config__)
class RoutineBase(RoutineParse, RoutineShow):
    source: Builder
    circuit: AnalogCircuit
    params: Params

    def __str__(self):
        out = self.__class__.__name__ + "\n"
        out = out + str(self.circuit)
        out = out + "\n---------------------\n"
        out = out + str(self.params)
        return out


@dataclass(frozen=True, config=__pydantic_dataclass_config__)
class Routine(RoutineBase):
    """Result of parsing a completed Builder string."""

    @property
    def braket(self) -> "BraketServiceOptions":
        from .braket import BraketServiceOptions

        return BraketServiceOptions(self.source, self.circuit, self.params)

    @property
    def quera(self) -> "QuEraServiceOptions":
        from .quera import QuEraServiceOptions

        return QuEraServiceOptions(self.source, self.circuit, self.params)

    @property
    def bloqade(self) -> "BloqadeServiceOptions":
        from .bloqade import BloqadeServiceOptions

        return BloqadeServiceOptions(self.source, self.circuit, self.params)



from beartype.typing import Union, TYPE_CHECKING
import bloqade.ir as ir
from bloqade.visualization import display_builder


if TYPE_CHECKING:
    from bloqade.ir import AtomArrangement, ParallelRegister, Sequence
    from bloqade.ir.analog_circuit import AnalogCircuit
    from bloqade.ir.routine.base import Routine
    from bloqade.builder.base import Builder


class ParseRegister:
    def parse_register(self: "Builder") -> Union["AtomArrangement", "ParallelRegister"]:
        """Parse the arrangement of atoms of the program."""
        from bloqade.builder.parse.builder import Parser

        return Parser().parse_register(self)


class ParseSequence:
    def parse_sequence(self: "Builder") -> "Sequence":
        """Parse the pulse sequence part of the program."""
        from bloqade.builder.parse.builder import Parser

        return Parser().parse_sequence(self)


class ParseCircuit:
    def parse_circuit(self: "Builder") -> "AnalogCircuit":
        """Parse the analog circuit from the program."""
        from bloqade.builder.parse.builder import Parser

        return Parser().parse_circuit(self)


class ParseRoutine:
    def parse(self: "Builder") -> "Routine":
        """Parse the program to return a Routine object."""
        from bloqade.builder.parse.builder import Parser

        return Parser().parse(self)


class Parse(ParseRegister, ParseSequence, ParseCircuit, ParseRoutine):
    @property
    def n_atoms(self: "Builder"):
        """Return the number of atoms in the program."""

        register = self.parse_register()

        if isinstance(register, ir.location.ParallelRegister):
            return register._register.n_atoms
        else:
            return register.n_atoms


class Show:
    def show(self, batch_id: int = 0):
        display_builder(self, batch_id)
